//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#define INCLUDE_GUARD_NAME "__HG_INFO_H__"

#ifdef _MSC_VER
#include <io.h>
#define W_OK 2

#define popen _popen
#define pclose _pclose
#define strdup _strdup
#endif

typedef enum
{
    false = 0,
    true = 1
} bool;

typedef struct DefinitionS
{
    char *Key;
    char *Value;
    bool IsInteger;

    struct DefinitionS *next;
} Definition;

static char *RevStr;
static int RevNum;

static Definition *Root;

static size_t ExecCmd(const char *command, char *output, size_t outsize);
static Definition *AddDef(const char *key, const char *cmd, bool asInteger, char **copy);
static Definition *AddDefParent(const char *key, const char *fmt, char **copy);
static Definition *AddDefParentInt(const char *key, const char *fmt, int *copy);
static void WriteOut(const char *outname);

size_t ExecCmd(const char *command, char *output, size_t outsize)
{
    FILE *f;
    size_t r;

    f = popen(command, "r");
    if (f == NULL)
    {
        fprintf(stderr, "error opening process %d: %s\n", errno, strerror(errno));
        exit(1);
    }

    r = fread(output, sizeof(char), outsize, f);
    if (ferror(f))
    {
        fprintf(stderr, "error reading from process %d: %s\n", errno, strerror(errno));
        exit(1);
    }

    output[r] = 0;

    pclose(f);

    return r;
}

Definition *AddDef(const char *key, const char *cmd, bool asInteger, char **copy)
{
    char *buf;
    Definition *def;

    buf = malloc(1024);
    ExecCmd(cmd, buf, 1024);

    def = malloc(sizeof(Definition));
    def->Key = strdup(key);
    def->Value = buf;
    def->IsInteger = asInteger;
    def->next = NULL;

    if (Root != NULL)
    {
        def->next = Root;
    }

    if (copy)
    {
        *copy = buf;
    }

    Root = def;
    return def;
}

Definition *AddDefParent(const char *key, const char *fmt, char **copy)
{
    char cmd[1024];
    sprintf(cmd, "hg parent --template \"%s\"", fmt);

    return AddDef(key, cmd, false, copy);
}

Definition *AddDefParentInt(const char *key, const char *fmt, int *copy)
{
    char cmd[1024];
    Definition *def;

    sprintf(cmd, "hg parent --template \"%s\"", fmt);
    def = AddDef(key, cmd, true, NULL);

    if (copy)
    {
        *copy = atoi(def->Value);
    }

    return def;
}

void WriteOut(const char *outname)
{
    char buf[10240];
    FILE *f;
    Definition *def;

    memset(buf, 0, sizeof(buf));

    sprintf(buf + strlen(buf), "// Automatically generated by hginfo tool.\n");
    sprintf(buf + strlen(buf), "#ifndef %s\n", INCLUDE_GUARD_NAME);
    sprintf(buf + strlen(buf), "#define %s\n", INCLUDE_GUARD_NAME);
    sprintf(buf + strlen(buf), "\n");

    for (def = Root; def != NULL; def = def->next)
    {
        sprintf(buf + strlen(buf), "#define %s ", def->Key);

        if (def->IsInteger)
        {
            sprintf(buf + strlen(buf), "%s", def->Value);
        }
        else
        {
            sprintf(buf + strlen(buf), "\"%s\"", def->Value);
        }

        sprintf(buf + strlen(buf), "\n", def->Value);
    }

    sprintf(buf + strlen(buf), "\n");

    sprintf(buf + strlen(buf), "#endif // %s\n", INCLUDE_GUARD_NAME);

    f = fopen(outname, "w");

    fwrite(buf, sizeof(char), strlen(buf), f);

    fclose(f);

    printf("Updated to hg revision %s (%d)\n", RevStr, RevNum);
}

int main(int argc, char **argv) {
    const char *outname;
    char revbuf[255];

    memset(revbuf, 0, sizeof(revbuf));

    if (argc < 2)
    {
        printf("syntax: %s <output path>\n", argv[0]);
        return 1;
    }

    outname = argv[1];

    AddDefParentInt("HG_REVISION", "{rev}", &RevNum);
    AddDefParent("HG_REVISION_STR", "{rev}", &RevStr);
    AddDefParent("HG_COMMIT", "{node|short}", NULL);
    AddDefParent("HG_COMMIT_LONG", "{node}", NULL);
    AddDefParent("HG_DATE", "{date|rfc822date}", NULL);

    WriteOut(outname);

    return 0;
}
