//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// Copyright 1993-1997 id Software
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// DESCRIPTION:
//	Handles key bindings..
//
//-----------------------------------------------------------------------------

#include <stdlib.h>
#include <string.h>
#include "c_binds.h"
#include "c_cmds.h"
#include "c_console.h"
#include "m_misc.h"
#include "i_system.h"
#include "m_alloc.h"

typedef struct bind_s {
    char *key;
    char *cmd;

    struct bind_s *next;
} bind_t;

bind_t *FirstBind = NULL;

static bind_t *FindBind(const char *key);

bind_t *FindBind(const char *key)
{
    bind_t *bind;

    for (bind = FirstBind; bind; bind = bind->next)
    {
        if (stricmp(bind->key, key) == 0)
        {
            return bind;
        }
    }

    return NULL;
}

boolean C_IsKeyBound(char *key)
{
    return FindBind(key) != NULL;
}

boolean C_BindsResponder(event_t *event)
{
    bind_t *bind;
    char *keyname;
    char *plus;
    boolean up;
    size_t i;

    up = (event->type == ev_keyup);

    keyname = (char *)I_MapKeyToName(event);
    if (!keyname)
        return false;

    keyname = strdup(keyname);
    for (i = 0; i < strlen(keyname); i++)
        keyname[i] = tolower(keyname[i]);

    bind = FindBind(keyname);
    if (!bind)
        return false;

    plus = strchr(bind->cmd, '+');

    if (plus && up)
        *plus = '-';

    if (plus || (!plus && !up))
        C_ExecuteMultiple(bind->cmd);

    if (plus && up)
        *plus = '+';

    M_Free(keyname);

    return true;
}

void C_RegisterBinding(char *key, char *cmd)
{
    bind_t *bind;

    bind = (bind_t *)M_Malloc(sizeof(bind_t));
    memset(bind, 0, sizeof(bind_t));

    bind->key = strdup(key);
    bind->cmd = strdup(cmd);

    C_RemoveBinding(key);

    bind->next = FirstBind;
    FirstBind = bind;
}

void C_RemoveBinding(char *key)
{
    bind_t *bind = NULL, *prev = NULL;
    
    for (bind = FirstBind; bind; bind = bind->next)
    {
        if (stricmp(bind->key, key) == 0)
        {
            if (prev) prev->next = bind->next;
            if (bind == FirstBind) FirstBind = bind->next;

            M_Free(bind);
            break;
        }

        prev = bind;
    }
}

char *C_GetBinding(char *key)
{
    bind_t *bind;

    bind = FindBind(key);
    if (bind)
        return bind->cmd;
    
    return NULL;
}

void C_SaveBindings(FILE *f)
{
    bind_t *bind;

    for (bind = FirstBind; bind; bind = bind->next)
    {
        fprintf(f, "bind \"%s\" \"%s\"\n", bind->key, bind->cmd);
    }
}
