//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// Copyright 1993-1997 id Software
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// DESCRIPTION:
//	Console command registration, registration, and dispatching.
//  Also some base commands.
//
//-----------------------------------------------------------------------------

#include <stdlib.h>
#include <string.h>
#include "c_console.h"
#include "c_cmds.h"
#include "c_binds.h"
#include "i_system.h"
#include "version.h"
#include "i_video.h"
#include "c_cvar.h"
#include "m_misc.h"
#include "z_zone.h"
#include "m_alloc.h"
#include "hginfo.h"
#include "d_main.h"
#include "doomdef.h"
#include "doomstat.h"
#include "m_menu.h"
#include "v_video.h"

typedef struct action_s {
    char *name;
    char *def;
    boolean down;
    int flag;

    struct action_s *next;
} action_t;

#define MAX_STRING_CHARS 1024
#define MAX_STRING_TOKENS 80
#define	MAX_TOKEN_CHARS 128

CommandFunction *FirstCommand = NULL;
action_t *FirstAction = NULL;

static char *C_MacroExpandString(char *text);
static void C_RegisterAction(const char *name, const char *def, int flag);
static action_t *C_FindAction(char *name);

int ActionFlags = 0;

static void ExitCmd(int argc, char **argv)
{
    I_Quit();
}

static void VersionCmd(int argc, char **argv)
{
    Printf("%s %s r%s\n", PACKAGE_NAME, PACKAGE_VERSION, HG_REVISION_STR);
    Printf("Built: %s %s\n", __DATE__, __TIME__);
}

static void EchoCmd(int argc, char **argv)
{
    int i;

    for (i = 1; i < argc; i++)
    {
        Printf("%s", argv[i]);

        if (i < argc - 1)
        {
            Printf(" ");
        }
    }

    Printf("\n");
}

static void ScreenshotCmd(int argc, char **argv)
{
    I_Screenshot();
}

static void SetCmd(int argc, char **argv)
{
    char *name, *value;

    if (argc < 3)
    {
        Printf("Syntax: %s <name> <value>\n", argv[0]);
        return;
    }

    name = argv[1];
    value = argv[2];

    SetCvar(name, value);
}

static void ToggleCmd(int argc, char **argv)
{
    char val[64];
    float newval;
    char *name;
    cvar_t *cvar;

    if (argc < 2)
    {
        Printf("Syntax: %s <name>\n", argv[0]);
        return;
    }

    name = argv[1];

    cvar = FindCvar(name);
    if (cvar == NULL)
    {
        Printf("CVar \"%s\" is not set\n", name);
        return;
    }

    newval = (float)!cvar->value;

    val[0] = '\0';
    sprintf(val, "%f", newval);

    SetCvar(name, val);
}

static void ExecCmd(int argc, char **argv)
{
    FILE *f;
    char cmd[256], *end, *fn;

    if (argc < 2)
    {
        Printf("Syntax: %s <file>\n", argv[0]);
        return;
    }

    fn = strdup(argv[1]);

    if ((f = fopen(fn, "r")))
    {
        while (fgets(cmd, 256, f))
        {
            end = cmd + strlen(cmd) - 1;
            if (*end == '\n')
                *end = '\0';
            C_ExecuteMultiple(cmd);
        }

        if (!feof(f))
        {
            Printf("Error reading from %s\n", fn);
        }
        else
        {
            Printf("Executed file %s\n", fn);
        }

        fclose(f);
    }
    else
    {
        Printf("Failed to open file %s", fn);
    }

    M_Free(fn);
}

static void DumpHeapCmd(int argc, char **argv)
{
    if (argc < 2)
    {
        Z_DumpHeap(0, 2147483647);
    }
    else
    {
        FILE *f;
        f = fopen(argv[1], "w");
        if (!f)
        {
            Printf("Failed to open %s for writing\n", argv[1]);
            return;
        }
        Z_FileDumpHeap(f);
        fclose(f);
        Printf("Wrote heap contents to %s\n", argv[1]);
    }
}

static void WriteHeapCmd(int argc, char **argv)
{
    if (argc < 2)
    {
        Printf("syntax: %s <filename>\n", argv[0]);
    }
    else
    {
        FILE *f;
        f = fopen(argv[1], "w");
        if (!f)
        {
            Printf("Failed to open %s for writing\n", argv[1]);
            return;
        }
        Z_WriteHeap(f);
        fclose(f);
        Printf("Wrote heap contents to %s\n", argv[1]);
    }
}

static void ErrorCmd(int argc, char **argv)
{
    char s[512];
    int i;

    memset(s, 0, 512);

    for (i = 1; i < argc; i++)
    {
        sprintf(s + strlen(s), "%s ", argv[i]);
    }

    I_Error(s);
}

static void ExecAction(int argc, char **argv)
{
    boolean down;
    char *name;
    action_t *act;

    down = argv[0][0] == '+';
    name = strdup(argv[0] + 1);

    act = C_FindAction(name);
    if (act == NULL)
    {
        Printf("Attempt to handle invalid action %s!\n", name);
        M_Free(name);
        return;
    }

    act->down = down;

    if (down)
    {
        ActionFlags |= act->flag;
    }
    else
    {
        ActionFlags &= ~act->flag;
    }

    M_Free(name);
}

static void DumpActionsCmd(int argc, char **argv)
{
    action_t *act;

    Printf("Current actions: %08x\n", ActionFlags);
    for (act = FirstAction; act; act = act->next)
    {
        Printf("%s\n", act->name);
        Printf("    Down: %s\n", act->down ? "yes" : "no");
        Printf("    Default key: %s\n", act->def);
        Printf("    Flag: %08x\n", act->flag);
    }
}

static void BindCmd(int argc, char **argv)
{
    char *key;
    char *cmd;

    if (argc < 2)
    {
        Printf("Syntax: %s <key> [command]\n", argv[0]);
        return;
    }

    key = argv[1];

    if (argc < 3)
    {
        char *bind = C_GetBinding(key);
        if (bind)
        {
            Printf("%s is bound to: %s\n", key, bind);
        }
        else
        {
            Printf("%s is not bound\n", key);
        }

        return;
    }

    cmd = argv[2];

    C_RegisterBinding(key, cmd);
}

static void UnbindCmd(int argc, char **argv)
{
    char *key;

    if (argc < 2)
    {
        Printf("Syntax: %s <key>\n", argv[0]);
        return;
    }

    key = argv[1];

    C_RemoveBinding(key);
}

static void EndGameCmd(int argc, char **argv) 
{
    if (gamestate == GS_DEMOSCREEN)
    {
        Printf("You must be in a game to use %s!\n", argv[0]);
        return;
    }

    if (Con_IsOpen())
        Con_Toggle();
    M_ClearMenus();
    D_StartTitle();
}

static void ToggleConsoleCmd(int argc, char **argv)
{
    Con_Toggle();
}

double gammalevel = 1;
static void GammaCmd(int argc, char **argv)
{
    if (argc < 2)
    {
        Printf("gamma is %f\n", gammalevel);
    }
    else
    {
        double newlevel;

        newlevel = atof(argv[1]);

        gammalevel = newlevel;
        V_CalculateGamma();
    }
}

extern int SetWeaponSlot;
static void SlotCmd(int argc, char **argv)
{
    if (argc < 2)
    {
        Printf("syntax: %s <number>\n", argv[0]);
    }
    else
    {
        SetWeaponSlot = atoi(argv[1]);
        if (SetWeaponSlot < 0 || SetWeaponSlot > 9)
            SetWeaponSlot = -1;
    }
}

extern void SndReinitCmd(int argc, char **argv);
extern void ListFmodDevicesCmd(int argc, char **argv);
extern void FmodStatusCmd(int argc, char **argv);
extern void DemoStopCmd(int argc, char **argv);

void C_InitCmds()
{
    C_RegisterCommand("exit", ExitCmd);
    C_RegisterCommand("quit", ExitCmd);
    C_RegisterCommand("version", VersionCmd);
    C_RegisterCommand("echo", EchoCmd);
    C_RegisterCommand("screenshot", ScreenshotCmd);
    C_RegisterCommand("set", SetCmd);
    C_RegisterCommand("toggle", ToggleCmd);
    C_RegisterCommand("exec", ExecCmd);
    C_RegisterCommand("dumpheap", DumpHeapCmd);
    C_RegisterCommand("writeheap", WriteHeapCmd);
    C_RegisterCommand("error", ErrorCmd);
    C_RegisterCommand("dumpactions", DumpActionsCmd);
    C_RegisterCommand("bind", BindCmd);
    C_RegisterCommand("unbind", UnbindCmd);
    C_RegisterCommand("endgame", EndGameCmd);
    C_RegisterCommand("toggleconsole", ToggleConsoleCmd);
    C_RegisterCommand("gamma", GammaCmd);
    C_RegisterCommand("slot", SlotCmd);
    C_RegisterCommand("demo_stop", DemoStopCmd);
    C_RegisterCommand("snd_reinit", SndReinitCmd);
    C_RegisterCommand("snd_listdevices", ListFmodDevicesCmd);
    C_RegisterCommand("snd_status", FmodStatusCmd);

    C_RegisterAction("forward", "up", ACTION_FORWARD);
    C_RegisterAction("backward", "down", ACTION_BACKWARD);
    C_RegisterAction("turnleft", "left", ACTION_TURNLEFT);
    C_RegisterAction("turnright", "right", ACTION_TURNRIGHT);
    C_RegisterAction("fire", "ctrl", ACTION_FIRE);
    C_RegisterAction("use", "space", ACTION_USE);
    C_RegisterAction("straferight", ".", ACTION_STRAFERIGHT);
    C_RegisterAction("strafeleft", ",", ACTION_STRAFELEFT);
    C_RegisterAction("speed", "shift", ACTION_SPEED);
    C_RegisterAction("strafe", "alt", ACTION_STRAFE);
    C_RegisterAction("lookup", "pgup", ACTION_LOOKUP);
    C_RegisterAction("lookdown", "pgdn", ACTION_LOOKDOWN);
    C_RegisterAction("lookcenter", "backspace", ACTION_LOOKCENTER);
    C_RegisterAction("jump", "/", ACTION_JUMP);
}

void C_RegisterCommand(const char *name, CommandCallback func)
{
    CommandFunction *cmd;
    int namesize;

    namesize = strlen(name);

    cmd = (CommandFunction *)M_Malloc(sizeof(CommandFunction));
    cmd->name = (char *)M_Malloc(namesize + 1);
    cmd->func = func;
    cmd->next = NULL;

    memset(cmd->name, 0, namesize + 1);
    memcpy(cmd->name, name, namesize);

    cmd->next = FirstCommand;
    FirstCommand = cmd;
}

void C_RegisterAction(const char *name, const char *def, int flag)
{
    action_t *key;
    char plusname[64];
    char minusname[64];

    memset(plusname, 0, 64);
    memset(minusname, 0, 64);

    key = (action_t *)M_Malloc(sizeof(action_t));
    memset(key, 0, sizeof(action_t));
    key->name = strdup(name);
    key->def = strdup(def);
    key->down = false;
    key->flag = flag;
    
    key->next = FirstAction;
    FirstAction = key;

    sprintf(plusname, "+%s", name);
    sprintf(minusname, "-%s", name);

    C_RegisterCommand(plusname, ExecAction);
    C_RegisterCommand(minusname, ExecAction);

    C_RegisterBinding((char *)def, plusname);
}

action_t *C_FindAction(char *name)
{
    action_t *key;

    for (key = FirstAction; key; key = key->next)
    {
        if (stricmp(key->name, name) == 0)
        {
            return key;
        }
    }

    return NULL;
}

char *COM_Parse(char **data_p)
{
    static boolean firstrun = true;
    static char com_token[MAX_TOKEN_CHARS];

    int c, len;
    char *data;

    if (firstrun)
    {
        firstrun = false;
        memset(com_token, 0, MAX_TOKEN_CHARS);
    }

    data = *data_p;
    len = 0;
    com_token[0] = 0;

    if (!data)
    {
        *data_p = NULL;
        return "";
    }

    // skip whitespace
skipwhite:
    while ((c = *data) <= ' ')
    {
        if (c == 0)
        {
            *data_p = NULL;
            return "";
        }
        data++;
    }

    // skip // comments
    if (c == '/' && data[1] == '/')
    {
        while (*data && *data != '\n')
            data++;
        goto skipwhite;
    }

    // handle quoted strings
    if (c == '"')
    {
        data++;
        while (1)
        {
            c = *data++;
            if (c == '"' || !c)
            {
                com_token[len] = 0;
                *data_p = data;
                return com_token;
            }
            if (len < MAX_TOKEN_CHARS)
            {
                com_token[len] = c;
                len++;
            }
        }
    }

    // parse regular word
    do
    {
        if (len < MAX_TOKEN_CHARS)
        {
            com_token[len] = c;
            len++;
        }

        data++;
        c = *data;
    } while (c > 32);

    if (len == MAX_TOKEN_CHARS)
    {
        len = 0;
    }

    com_token[len] = 0;

    *data_p = data;
    return com_token;
}

char *C_MacroExpandString(char *text)
{
    int i, j, count, len;
    boolean inquote;
    char *scan;
    static char expanded[MAX_STRING_CHARS];
    char temporary[MAX_STRING_CHARS];
    char *token, *start;

    inquote = false;
    scan = text;

    len = strlen(text);
    if (len >= MAX_STRING_CHARS)
    {
        Printf("Line exceeded %i chars, discarded.\n", MAX_STRING_CHARS);
        return NULL;
    }

    count = 0;

    for (i = 0; i < len; i++)
    {
        if (scan[i] == '"')
            inquote ^= 1;
        if (inquote)
            continue;
        if (scan[i] != '$')
            continue;

        start = scan + i + 1;
        token = COM_Parse(&start);

        if (!start)
            continue;

        token = CvarString(token);

        j = strlen(token);
        len += j;

        if (len >= MAX_STRING_CHARS)
        {
            Printf("Expanded line exceeded %i chars, discarded.\n", MAX_STRING_CHARS);
            return NULL;
        }

        strncpy(temporary, scan, i);
        strcpy(temporary + i, token);
        strcpy(temporary + i + j, start);

        strcpy(expanded, temporary);

        scan = expanded;
        
        i--;

        if (++count == 100)
        {
            Printf("Macro expansion loop, discarded.\n");
            return NULL;
        }
    }

    if (inquote)
    {
        Printf("Line has unmatched quote, discarded.\n");
        return NULL;
    }

    return scan;
}

int C_TokeniseString(char *text, char ***_argv, boolean expandMacros)
{
    int i;
    char *com_token;
    static int argc = 0;
    static char *argv[MAX_STRING_TOKENS];
    static char args[MAX_STRING_CHARS];

    // clear args from last string
    for (i = 0; i < argc; i++)
    {
        M_Free(argv[i]);
    }

    argc = 0;

    if (expandMacros)
        text = C_MacroExpandString(text);

    if (!text)
        return 0;

    while (1)
    {
        while (*text && *text <= ' ' && *text != '\n')
        {
            text++;
        }

        if (*text == '\n')
        {
            text++;
            break;
        }

        if (!*text)
        {
            *_argv = argv;
            return argc;
        }

        if (argc == 1)
        {
            int l;
            strcpy(args, text);

            l = strlen(args) - 1;
            for (; l >= 0; l--)
            {
                if (args[l] <= ' ')
                    args[l] = 0;
                else
                    break;
            }
        }

        com_token = COM_Parse(&text);
        if (!text)
            return 0;

        if (argc < MAX_STRING_TOKENS)
        {
            argv[argc] = (char *)M_Calloc(strlen(com_token) + 1, sizeof(char));
            memcpy(argv[argc], com_token, strlen(com_token));
            argc++;
        }
    }

    *_argv = argv;
    return argc;
}

void C_ExecuteString(char *text)
{
    int argc;
    char **argv;
    CommandFunction *cmd;
    cvar_t *cvar;

    argc = C_TokeniseString(text, &argv, true);
    if (argc == 0)
    {
        return;
    }

    cmd = FindConsoleCommand(argv[0]);
    if (cmd)
    {
        cmd->func(argc, argv);
        return;
    }

    cvar = FindCvar(argv[0]);
    if (cvar)
    {
        if (argc > 1)
        {
            SetCvar(argv[0], argv[1]);
        }
        else
        {
            Printf("%s is set to \"%s\"\n", cvar->name, cvar->string);
        }

        return;
    }

    Printf("Unknown command \"%s\"\n", argv[0]);
}

void C_ExecuteMultiple(char *text)
{
    char *b;
    int more;

    while (*text) 
    {
        b = text;

        while (*b != ';' && *b != '\0')
        {
            if (*b == '"')
            {
                b++;
                while (*b != '"' && *b != 0)
                {
                    b++;
                }
            }
            b++;
        }

        if (*b == ';')
        {
            *b = 0;
            more = 1;
        }
        else
        {
            more = 0;
        }

        C_ExecuteString(text);

        if (more)
        {
            *b = ';';
        }

        text = b + more;
    }
}

CommandFunction *FindConsoleCommand(const char *name)
{
    CommandFunction *cmd;

    for (cmd = FirstCommand; cmd; cmd = cmd->next)
    {
        if (strcmp(cmd->name, name) == 0)
        {
            return cmd;
        }
    }

    return NULL;
}
