//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// Copyright 1993-1997 id Software
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
//
// 3. Neither the name of the copyright holder nor the names of its
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// DESCRIPTION:
//  Console. Most of this code was "borrowed" from Quake 2...
//
//-----------------------------------------------------------------------------

#include <stdarg.h>
#include <stdlib.h>
#include <ctype.h>
#include "c_binds.h"
#include "c_console.h"
#include "c_cmds.h"
#include "c_cvar.h"
#include "hu_lib.h"
#include "doomstat.h"
#include "hu_stuff.h"
#include "i_system.h"
#include "multires.h"
#include "m_misc.h"
#include "z_zone.h"
#include "w_wad.h"
#include "m_alloc.h"
#include "version.h"
#include "hginfo.h"
#include "i_video.h"

enum
{
    ConOpen,
    ConClosed
};

#define CON_TEXTSIZE 32700
#define CON_INPUTBUFSIZE 512
#define NUM_CON_TIMES 4
#define NUM_CON_HISTORY 32

char ConText[CON_TEXTSIZE];
int ConTimes[NUM_CON_TIMES];
char ConInput[CON_INPUTBUFSIZE];

// Console height is SCREENHEIGHT/con_heightdiv
cvar_t *con_heightdiv;
cvar_t *developer;
cvar_t *con_staytime;
extern cvar_t *showmessages;

boolean ConInitialised;

// Line where next message will be printed
static int ConCurrent;

// Offset in current line for next print
static int ConX;

// Bottom of console displays this line
static int ConDisplay;

// Character across screen
static int ConLineWidth;

// Total lines that can fit in the console buffer
static int ConTotalLines;

// State of the console
static int ConsoleState;

// Last time a character was appended to the input
// (This is used to reset the cursor flashing after typing a character)
static int ConTypeTime;

// Position of the input cursor
static unsigned ConCursor;

// Console input history
char *ConHistory[NUM_CON_HISTORY + 1];
int ConHistoryPosition;

static struct {
    int Height;
} ConDrawVars;

static byte *ConBackground;

patch_t *ConFont[256];

static void Con_LoadFont();
static void Con_CheckResize();
static void Con_GenerateBackground();
static void Con_LineFeed();
static void Con_DrawNotify();
static void Con_DrawInput();
static void Con_DrawVersion();
static void Con_PushHistory();

void Con_LoadFont()
{
    int i;
    char buffer[9];

    for (i = 0; i < 256; i++)
    {
        if (i == 0x0F)
            strcpy(buffer, "ALTCONUP");
        else
            sprintf(buffer, "ALCONF%02X", i);

        ConFont[i] = (patch_t *)W_CacheLumpName(buffer, PU_STATIC);
    }
}

void Con_Init()
{
    Con_LoadFont();

    memset(ConInput, 0, CON_INPUTBUFSIZE);

    ConsoleState = ConClosed;
    ConLineWidth = -1;
    ConCursor = 0;
    ConHistoryPosition = 0;

    Con_CheckResize();

    ConInitialised = true;

    ConTypeTime = I_GetTime();
}

void Con_Ticker()
{
    //ActionFlags = 0;
};

void Con_Toggle()
{
    if (ConsoleState == ConClosed)
    {
        ConsoleState = ConOpen;
        I_StartTyping();
    }
    else
    {
        I_StopTyping();
        ConsoleState = ConClosed;
    }

    menuactive = false;
}

boolean Con_IsOpen()
{
    return ConsoleState == ConOpen;
}

void Con_ResetNotify()
{
    int i;

    for (i = 0; i < NUM_CON_TIMES; i++)
    {
        ConTimes[i] = 0;
    }
}

void Con_CheckResize()
{
    int i, j, width, oldwidth, oldtotallines, numlines, numchars;
    char tbuf[CON_TEXTSIZE];

    width = (SCREENWIDTH / ConFont['H']->width);

    if (width == ConLineWidth)
    {
        return;
    }

    oldwidth = ConLineWidth;
    ConLineWidth = width;
    oldtotallines = ConTotalLines;
    ConTotalLines = ConLineWidth;
    numlines = oldtotallines;

    if (ConTotalLines < numlines)
    {
        numlines = ConTotalLines;
    }

    numchars = oldwidth;

    if (ConLineWidth < numchars)
    {
        numchars = ConLineWidth;
    }

    memcpy(tbuf, ConText, CON_TEXTSIZE);
    memset(ConText, ' ', CON_TEXTSIZE);

    for (i = 0; i < numlines; i++)
    {
        for (j = 0; j < numchars; j++)
        {
            ConText[(ConTotalLines - 1 - i) * ConLineWidth + j] = tbuf[((ConCurrent - i + oldtotallines) % oldtotallines) * oldwidth + j];
        }
    }

    ConCurrent = ConTotalLines - 1;
    ConDisplay = ConCurrent;

    Con_GenerateBackground();
}

void Con_GenerateBackground()
{
    int bgsize;

    if (ConBackground != NULL)
    {
        M_Free(ConBackground);
    }

    bgsize = SCREENWIDTH * SCREENHEIGHT;
    ConBackground = (byte *)M_Malloc(bgsize);
    memset(ConBackground, 0, bgsize);

    {
        // create a temp screen to render the titlepic
        screen_t tmpscreen;
        int lump;
        patch_t *titlepic;

        lump = W_CheckNumForName("CONBACK");
        if (lump == -1)
            lump = W_GetNumForName("TITLEPIC");
        titlepic = (patch_t *)W_CacheLumpNum(lump, PU_CACHE);

        I_AllocScreen(&tmpscreen, SCREENWIDTH, SCREENHEIGHT, 1);
        I_LockScreen(&tmpscreen);

        V_DrawPatchInDirect(0, 0, &tmpscreen, titlepic);
        V_GetBlock(0, 0, &tmpscreen, SCREENWIDTH, SCREENHEIGHT, ConBackground);

        I_UnlockScreen(&tmpscreen);
        I_FreeScreen(&tmpscreen);
    }

    {
        byte *palette = (byte *)W_CacheLumpName("PLAYPAL", PU_CACHE);
        int i;

        for (i = 0; i < bgsize; i++)
        {
            const int dimfactor = 6;
            byte r, g, b;
            byte oldcol, newcol;

            oldcol = ConBackground[i];

            V_PalleteIndexToColor(oldcol, 0, &r, &g, &b);

            // dim it
            r /= dimfactor;
            g /= dimfactor;
            b /= dimfactor;

            // find closest
            newcol = V_ClosestColor(r, g, b);

            ConBackground[i] = newcol;
        }
    }
}

void Con_LineFeed()
{
    ConX = 0;

    if (ConDisplay == ConCurrent)
    {
        ConDisplay++;
    }

    ConCurrent++;
    memset(&ConText[(ConCurrent % ConTotalLines) * ConLineWidth], ' ', ConLineWidth);
}

void Con_Draw()
{
    int i, y, x;
    int rows;
    char *text;
    int row;
    int lines;
    int ch;
    int condiv;
    condiv = 2;

    // reset heightdiv if invalid
    if (con_heightdiv->value < 1.0f)
    {
        SetCvar(con_heightdiv->name, "2.0");
    }

    if (ConsoleState == ConClosed)
    {
        Con_DrawNotify();
        return;
    }

    Con_ResetNotify();

    ConDrawVars.Height = (int)((float)SCREENHEIGHT / con_heightdiv->value);
    if (gamestate != GS_LEVEL)
        ConDrawVars.Height = SCREENHEIGHT;

    V_DrawBlock(0, 0 - (SCREENHEIGHT - ConDrawVars.Height), &screens[0], SCREENWIDTH, SCREENHEIGHT, ConBackground);

    ch = Con_CharHeight();

    lines = ConDrawVars.Height / ch;

    rows = lines;
    y = ConDrawVars.Height - (ch * 2);

    // draw arrows if scrolled up
    if (ConDisplay != ConCurrent)
    {
        const char *st = "\x0F";
        int len = Con_StringWidth(st);

        for (x = 0; x < ConLineWidth * len; x += len)
        {
            Con_DrawString(x, y + ch, st);
        }
    }

    row = ConDisplay;
    for (i = 0; i < rows; i++, y -= ch, row--)
    {
        if (row < 0)
        {
            break;
        }

        if (ConCurrent - row >= ConTotalLines)
        {
            break;
        }

        text = ConText + (row % ConTotalLines) * ConLineWidth;

        Con_DrawString(0, y, text);
    }

    Con_DrawVersion();
    Con_DrawInput();
}

void Con_DrawNotify()
{
    int v, ch;
    char *text;
    int i, time;

    if (!showmessages->value)
    {
        return;
    }

    ch = Con_CharHeight();
    v = 0;

    for (i = ConCurrent - NUM_CON_TIMES + 1; i <= ConCurrent; i++)
    {
        if (i < 0)
        {
            continue;
        }

        time = ConTimes[i % NUM_CON_TIMES];
        if (time == 0)
        {
            continue;
        }

        time = I_GetTime() - time;
        if (time > con_staytime->value * TICRATE)
        {
            continue;
        }

        text = ConText + (i % ConTotalLines) * ConLineWidth;

        Con_DrawString(0, v, text);

        v += ch;
    }
}

void Con_DrawInput()
{
    int y;
    int x;
    int cx, ch;
    int curx;
    char buf[CON_INPUTBUFSIZE + 16];
    const char *prefix = "\x11 ";

    // Insert space where cursor is
    memset(buf, 0, CON_INPUTBUFSIZE + 16);
    memcpy(buf, ConInput, CON_INPUTBUFSIZE);
    if (ConCursor != strlen(ConInput))
    {
        memcpy(buf + ConCursor + 1, ConInput + ConCursor, CON_INPUTBUFSIZE - ConCursor);
        buf[ConCursor] = ' ';
    }

    ch = Con_CharHeight();

    x = Con_StringWidth(prefix);

    y = ConDrawVars.Height - ch;

    cx = x + (Con_StringWidth(buf));

    Con_DrawString(0, y, prefix);
    Con_DrawString(x, y, buf);

    if (((I_GetTime() - ConTypeTime) / 12) % 2 == 0)
    {
        buf[ConCursor] = 0;
        curx = Con_StringWidth(buf) + Con_StringWidth(prefix);
        Con_DrawString(curx, y, "_");
    }
}

void Con_DrawString(int x, int y, const char *str)
{
    int i, cx, w, h;
    char *ptr;
    char c;
    int hui;

    ptr = (char *)str;

    cx = 0;
    i = 0;

    while ((c = *(ptr++)))
    {
        hui = (int)c;
        if (c == ' ')
        {
            cx += ConFont[' ']->width;
            continue;
        }


        w = ConFont[hui]->width;
        h = ConFont[hui]->height;
        if (cx + w > SCREENWIDTH || cx < 0 || y < 0 || y + h > SCREENHEIGHT)
        {
            break;
        }

        V_TranslationTable = V_TranslationTables + (CR_BLACK * 256);
        V_DrawPatchTranslated(x + cx + 1, y + 1, &screens[0], ConFont[hui]);
        V_DrawPatch(x + cx, y, &screens[0], ConFont[hui]);

        cx += w;
    }
}

void Con_DrawVersion()
{
    static char buf[64];
    static boolean inited = false;
    static int bufwidth;
    int inputwidth;
    int drawx;

    if (!inited)
    {
        inited = true;

        memset(buf, 0, sizeof(buf));
        sprintf(buf, "%s v%s r%s (%s)", PACKAGE_NAME, PACKAGE_VERSION, HG_REVISION_STR, __DATE__);

        bufwidth = Con_StringWidth(buf);
    }

    inputwidth = Con_StringWidth(ConInput) + Con_StringWidth("]  ");

    drawx = SCREENWIDTH - bufwidth;

    if (inputwidth < drawx)
    {
        Con_DrawString(drawx, ConDrawVars.Height - ConFont['H']->height, buf);
    }
}

int Con_StringWidth(const char *str)
{
    int sw, w, len;
    char *ptr;
    char c;

    ptr = (char *)str;

    len = 0;
    sw = 0;

    while ((c = *(ptr++)))
    {
        int hui = (int)c;

        w = ConFont[hui]->width;

        sw += w;
        len++;
    }

    return sw;
}

boolean Con_Responder(event_t *ev)
{
    char b[2];
    char c;
    int il;
    char *entry;

    b[1] = 0;

    if (ev->type == ev_keydown)
    {
        switch (ev->data1)
        {

        case KEY_BACKSPACE:
            if (ConsoleState != ConOpen) return false;

            il = strlen(ConInput);
            if (il < 0) return true;

            ConInput[il - 1] = '\0';
            ConTypeTime = I_GetTime();
            if (ConCursor > 0)
                ConCursor--;

            return true;

        case KEY_ENTER:
            if (ConsoleState != ConOpen) return false;

            // dispatch command
            Printf("\x11 %s\n", ConInput);
            C_ExecuteString(ConInput);
            Con_PushHistory();
            memset(ConInput, 0, CON_INPUTBUFSIZE);
            ConCursor = 0;
            ConHistoryPosition = 0;

            return true;

        case KEY_PGUP:
            if (ConsoleState != ConOpen) return false;

            ConDisplay -= 2;

            return true;

        case KEY_PGDN:
            if (ConsoleState != ConOpen) return false;

            ConDisplay += 2;

            if (ConDisplay > ConCurrent)
                ConDisplay = ConCurrent;

            return true;

        case KEY_LEFTARROW:
            if (ConsoleState != ConOpen) return false;

            if (ConCursor > 0)
                ConCursor--;

            return true;

        case KEY_RIGHTARROW:
            if (ConsoleState != ConOpen) return false;

            if (ConCursor < strlen(ConInput))
                ConCursor++;

            return true;

        case KEY_UPARROW:
            if (ConsoleState != ConOpen) return false;

            ConHistoryPosition++;
            if (ConHistoryPosition >= NUM_CON_HISTORY || (entry = ConHistory[ConHistoryPosition]) == NULL)
            {
                ConHistoryPosition--;
            }
            else
            {
                memset(ConInput, 0, CON_INPUTBUFSIZE);
                memcpy(ConInput, entry, strlen(entry));

                ConCursor = strlen(ConInput);
            }

            return true;

        case KEY_DOWNARROW:
            if (ConsoleState != ConOpen) return false;

            if (ConHistoryPosition == 1)
            {
                ConHistoryPosition--;
                memset(ConInput, 0, CON_INPUTBUFSIZE);
                ConCursor = 0;
            }
            else if (--ConHistoryPosition < 0 || (entry = ConHistory[ConHistoryPosition]) == NULL)
            {
                ConHistoryPosition++;
            }
            else
            {
                memset(ConInput, 0, CON_INPUTBUFSIZE);
                memcpy(ConInput, entry, strlen(entry));

                ConCursor = strlen(ConInput);
            }

            return true;

        default:
            break;
        }
    }

    // append to buffer
    if (ev->type == ev_type)
    {
        size_t inlen = strlen(ConInput);

        c = (char)ev->data1;

        if (c < 32 || c > 127)
            return true;

        if (ev->data3 == 41 || c == '`')
        {
            Con_Toggle();
            return true;
        }

        if (ConsoleState != ConOpen) return false;

        b[0] = c;

        if (ConCursor == inlen)
        {
            strncat(ConInput, b, CON_INPUTBUFSIZE);
        }
        else
        {
            memcpy(ConInput + ConCursor + 1, ConInput + ConCursor, inlen - ConCursor);
            ConInput[ConCursor] = c;
        }

        ConTypeTime = I_GetTime();
        ConCursor++;

        return true;
    }

    // inhibit binds if console is open
    if (ConsoleState == ConOpen)
        return true;

    // fall through to next responder
    return false;
}

void Con_PushHistory()
{
    unsigned i, num;
    char *buf;
    size_t len;
    boolean iswhitespace = true;

    len = strlen(ConInput);

    // don't bother pushing empty strings
    if (len < 1)
        return;

    // or blank ones
    for (i = 0; i < len; i++)
    {
        if (!isspace(ConInput[i]))
        {
            iswhitespace = false;
            break;
        }
    }

    if (iswhitespace)
        return;

    // or duplicates
    if (ConHistory[1] && strcmp(ConInput, ConHistory[1]) == 0)
        return;

    for (num = 1; num < NUM_CON_HISTORY; num++)
    {
        if (ConHistory[num] == NULL)
            break;
    }

    if (num  - 1 == NUM_CON_HISTORY)
    {
        M_Free(ConHistory[num]);
        ConHistory[num] = NULL;
        num--;
    }

    memmove(ConHistory + 1, ConHistory, num * sizeof(char *));

    buf = M_Calloc(len + 1, sizeof(char));
    memcpy(buf, ConInput, len);
    ConHistory[1] = buf;
}

void Con_PrintString(char *text)
{
    int y, l;
    char c;
    static boolean cr;

    if (!ConInitialised)
    {
        return;
    }

    while ((c = *text))
    {
        for (l = 0; l < ConLineWidth; l++)
        {
            if (text[l] <= ' ')
            {
                break;
            }
        }

        // word wrap
        if (l != ConLineWidth && (ConX + l > ConLineWidth))
        {
            ConX = 0;
        }

        text++;

        if (cr)
        {
            ConCurrent--;
            cr = false;
        }

        if (!ConX)
        {
            Con_LineFeed();

            if (ConCurrent >= 0)
            {
                ConTimes[ConCurrent % NUM_CON_TIMES] = I_GetTime();
            }
        }

        switch (c)
        {
        case '\n':
            ConX = 0;
            break;

        case '\r':
            ConX = 0;
            cr = true;
            break;

        default:
            y = ConCurrent % ConTotalLines;
            ConText[y * ConLineWidth + ConX] = c;
            ConX++;

            if (ConX >= ConLineWidth)
            {
                ConX = 0;
            }
           
            break;
        }
    }

    if (ConCurrent >= 0)
    {
        ConTimes[ConCurrent % NUM_CON_TIMES] = I_GetTime();
    }
}

void Printf(const char *fmt, ...)
{
    va_list args;
    char buf[1024];
   
    va_start (args, fmt);
    vsprintf (buf, fmt, args);
    va_end (args);

    Con_PrintString(buf);
    printf("%s", buf);
}

void DevPrintf(const char *fmt, ...)
{
    va_list args;
    char buf[1024];

    if (!developer->value)
        return;
   
    va_start (args, fmt);
    vsprintf (buf, fmt, args);
    va_end (args);

    Con_PrintString(buf);
    printf("%s", buf);
}

int Con_CharHeight()
{
    return ConFont['H']->height;
}
