//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// Copyright 1993-1997 id Software
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// DESCRIPTION:
//	Console variable registration and functions.
//
//-----------------------------------------------------------------------------

#include <stdlib.h>
#include <string.h>
#include "c_cvar.h"
#include "c_console.h"
#include "m_misc.h"
#include "m_alloc.h"

cvar_t *FirstCvar = NULL;

static cvar_t *SetCvarInternal(char *name, char *value, boolean force);

cvar_t *FindCvar(const char *name)
{
    cvar_t *v = FirstCvar;

    while (v)
    {
        if (stricmp(v->name, name) == 0)
        {
            return v;
        }

        v = v->next;
    }

    return NULL;
}

cvar_t *GetCvar(char *name, char *value, int flags)
{
    cvar_t *var;

    var = FindCvar(name);
    if (var)
    {
        var->flags |= flags;
        return var;
    }

    if (!value)
    {
        return NULL;
    }

    var = (cvar_t *)M_Malloc(sizeof(cvar_t));
    memset(var, 0, sizeof(cvar_t));
    var->name = strdup(name);
    var->string = strdup(value);
    var->value = (float)atof(var->string);
    var->modified = true;
    
    var->next = FirstCvar;
    FirstCvar = var;

    var->flags = flags;

    return var;
}

cvar_t *SetCvarInternal(char *name, char *value, boolean force)
{
    cvar_t *var;

    var = FindCvar(name);

    if (!var)
    {
        return GetCvar(name, value, 0);
    }

    if (!force)
    {
        if (var->flags & CVAR_NOSET)
        {
            Printf("%s is write protected.\n", name);
            return var;
        }
    }

    if (strcmp(value, var->string) == 0)
    {
        return var;
    }

    var->modified = true;

    M_Free(var->string);

    var->string = strdup(value);
    var->value = (float)atof(var->string);
    
    return var;
}

cvar_t *SetCvar(char *name, char *value)
{
    return SetCvarInternal(name, value, false);
}

cvar_t *ForceSetCvar(char *name, char *value)
{
    return SetCvarInternal(name, value, true);
}

char *CvarString(char *name)
{
    cvar_t *var;

    var = FindCvar(name);
    if (!var)
        return "";

    return var->string;
}

void C_SaveCvars(FILE *f)
{
    cvar_t *cvar;

    for (cvar = FirstCvar; cvar; cvar = cvar->next)
    {
        if (cvar->flags & CVAR_ARCHIVE)
        {
            fprintf(f, "set \"%s\" \"%s\"\n", cvar->name, cvar->string);
        }
    }
}

void C_CvarSet(cvar_t *cv, char *value)
{
    SetCvarInternal(cv->name, value, false);
}

void C_CvarSetFloat(cvar_t *cv, float value)
{
    char v[64];
    memset(v, 0, 64);

    sprintf(v, "%f", value);

    SetCvarInternal(cv->name, v, false);
}