//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// Copyright 1993-1997 id Software
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// DESCRIPTION:
//	IWAD detection code.
//
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "d_iwad.h"
#include "d_main.h"
#include "doomstat.h"
#include "c_console.h"
#include "i_system.h"
#include "md5.h"
#include "m_alloc.h"
#include "m_argv.h"
#include "w_wad.h"

#ifdef _MSC_VER
#include <io.h>
#define access _access
#define R_OK 4
#endif

#define NUM_IWADS 10
static IWadInfo iwadDefs[NUM_IWADS] = 
{
    // Chex Quest
    { 
        "Chex Quest",
        "doom",
        "doom",
        IWF_NONE,
        registered,

        // Required lumps
        {
            "E1M1", "W94_1", "POSSH0M0",
            NULL
        }
    },
        
    // Freedoom
    { 
        "Freedoom: Phase 1",
        "doom",
        "doom",
        IWF_FREEDOOM,
        registered,

        // Required lumps
        {
            "E1M1", "FREEDOOM",
            NULL
        }
    },

    // FreeDM
    { 
        "FreeDM",
        "doom2",
        "doom2",
        IWF_FREEDOOM,
        commercial,

        // Required lumps
        {
            "MAP01", "FREEDM",
            NULL
        }
    },

    // Freedoom 2
    { 
        "Freedoom: Phase 2",
        "doom2",
        "doom2",
        IWF_FREEDOOM,
        commercial,

        // Required lumps
        {
            "MAP01", "FREEDOOM",
            NULL
        }
    },

    // The Ultimate Doom
    {
        "The Ultimate Doom",
        "doom_ultimate",
        "doom",
        IWF_NONE,
        retail,

        // Required lumps
        {
            "E1M1", "E2M2", "E3M3", "DSPISTOL", "DPHOOF", "CYBRA1",
            "E4M2",
            NULL
        }
    },

    // Doom
    { 
        "Doom",
        "doom",
        "doom",
        IWF_NONE,
        registered,

        // Required lumps
        {
            "E1M1", "E2M2", "E3M3", "DSPISTOL", "DPHOOF", "CYBRA1",
            NULL
        }
    },

    // Doom (shareware)
    {
        "Doom (Shareware)",
        "doom_shareware",
        "doom",
        IWF_SHAREWARE,
        shareware,

        // Required lumps
        {
            "E1M1",
            NULL
        }
    },

    // TNT: Evilution
    {
        "TNT: Evilution",
        "tnt",
        "tnt",
        IWF_NONE,
        commercial,

        // Required lumps
        {
            "MAP01", "REDTNT2",
            NULL
        }
    },

    // The Plutonia Experiment
    {
        "The Plutonia Experiment",
        "plutonia",
        "plutonia",
        IWF_NONE,
        commercial,

        // Required lumps
        {
            "MAP01", "CAMO1",
            NULL
        }
    },

    // Doom 2
    { 
        "Doom II: Hell on Earth",
        "doom2",
        "doom2",
        IWF_NONE,
        commercial,

        // Required lumps
        {
            "MAP01",
            NULL
        }
    },
};

static const char *CandidateNames[] = 
{
    "doom.wad",
    "udoom.wad",
    "doom1.wad",
    "doom2.wad",
    "tnt.wad",
    "plutonia.wad",
    "freedoom1.wad",
    "freedoom2.wad",
    "freedm.wad",
    NULL
};

char *CurrentIwadFile;
const IWadInfo *CurrentIwad = NULL;

static void CheckWadForLumps();

static boolean CheckForCandidate(const char *name);
static boolean CheckCandidateInDirectory(const char *dir, const char *name);

void D_IdentifyIWAD()
{
    W_AddFile(CurrentIwadFile);

    // Check the WAD by lump
    // This is safe, as at this point the only WAD loaded is the IWAD.
    {
        int i, j;
        for (i = 0; i < NUM_IWADS; i++)
        {
            const IWadInfo *info = &iwadDefs[i];
            int total = 0, matched = 0;

            for (j = 0; j < NUM_IWAD_LUMPS; j++)
            {
                const char *lumpname = info->RequiredLumps[j];

                if (lumpname == NULL)
                    break;

                total++;
                
                if (W_CheckNumForName((char *)lumpname) != -1)
                    matched++;
            }

            if (total == matched)
            {
                CurrentIwad = info;	
                goto found_iwad;
            }
        }
    }

    // Nuke all mounted WADs so we can start again fresh
    W_Clear();

    // Welp.
    if (CurrentIwad == NULL)
    {
        gamemode = indetermined;
        Printf("D_IdentifyIWAD: Failed to identify IWAD %s\n", CurrentIwadFile);
        return;
    }

found_iwad:
    Printf("Playing %s (%s)\n", CurrentIwad->Name, CurrentIwad->Id);

    gamemode = (GameMode_t)CurrentIwad->GameMode;
}

boolean D_FindIWAD()
{
    int p, i = 0;

    // First check if an IWAD is specified on the command line
    if ((p = M_CheckParm("-iwad")))
    {
        CurrentIwadFile = strdup(myargv[p + 1]);
        return true;
    }

    while (CandidateNames[i])
    {
        const char *name = CandidateNames[i];

        if (CheckForCandidate(name))
        {
            return true;
        }

        i++;
    }

    // We failed to find an IWAD.
    // But a loaded PWAD might still work!
    return false;
}

boolean CheckCandidateInDirectory(const char *dir, const char *name)
{
    FILE *f;
    char buf[512];
    sprintf(buf, "%s/%s", dir, name);

    // Dumb, but access(buf, R_OK) doesn't work on Windows for some reason, so
    // we try to open the file instead.
    f = fopen(buf, "rb");
    if (!f) return false;
    fclose(f);
    CurrentIwadFile = strdup(buf);
    return true;
}

boolean CheckForCandidate(const char *name)
{
    char *doomwaddir = getenv("DOOMWADDIR");

    if (CheckCandidateInDirectory(".", name)) return true;
    if (CheckCandidateInDirectory(I_GetExeDir(), name)) return true;
    if (doomwaddir && CheckCandidateInDirectory(doomwaddir, name)) return true;

    return false;
}
