//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// Copyright 1993-1997 id Software
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
//-----------------------------------------------------------------------------

#include <fmod.h>
#include <fmod_errors.h>
#include "i_sound.h"
#include "i_system.h"
#include "c_console.h"
#include "m_alloc.h"
#include "m_stream.h"
#include "w_wad.h"
#include "c_cvar.h"
#include "m_argv.h"
#include "v_video.h"
#include "st_stuff.h"
#include "m_fixed.h"

#if defined(_MSC_VER) || defined(__MINGW32__)
#include <malloc.h>
#else
#include <alloca.h>
#endif

// Needed for stats counting
typedef struct 
{
    boolean IsStream;
} FmodUserData;

struct 
{
    int StreamBytes;
    int SampleBytes;
} FmodStats;

static FMOD_SYSTEM *System;
static FMOD_CHANNELGROUP *SoundsChannels, *MusicChannels;
static FMOD_CHANNEL *MusicChannel;
static FMOD_SOUND *ActiveMusic;

// qmus2mid.c
extern int convert(const char *mus, const char *mid, int nodisplay, int div, int size, int nocomp, int *ow);

static FMOD_SOUND *LoadFmodSound(const char *name, FMOD_CHANNEL **chan, FMOD_MODE mode);
static byte *ConvertMusToMidi(int lump, unsigned *outSize);

static FMOD_RESULT F_CALLBACK FmodOpenFile(const char *name, unsigned *filesize, void **handle, void *userdata);
static FMOD_RESULT F_CALLBACK FmodCloseFile(void *handle, void *userdata);
static FMOD_RESULT F_CALLBACK FmodReadFile(void *handle, void *buffer, unsigned sizebytes, unsigned *bytesread, void *userdata);
static FMOD_RESULT F_CALLBACK FmodSeekFile(void *handle, unsigned pos, void *userdata);
static void AddFmodCallbacks(FMOD_CREATESOUNDEXINFO *exinfo);

extern cvar_t *snd_numchannels,
    *snd_musvolume;

cvar_t *snd_device,
    *snd_ignorepadding,
    *snd_drawoverlay;

static boolean nosound = false;

static void I_FmodError(FMOD_RESULT result)
{
    const char *str;

    str = FMOD_ErrorString(result);

    I_Error("An FMOD error has occured:\n%s", str);
}

static void I_CheckFmodError(FMOD_RESULT result)
{
    const char *str;

    if (result != FMOD_OK)
    {
        I_ShutdownSound();
        nosound = true;
        str = FMOD_ErrorString(result);
        Printf("An FMOD error has occured: %s\n", str);
    }
}

void ListFmodDevicesCmd(int argc, char **argv)
{
    int num;
    FMOD_RESULT result;
    int i, current;

    if (!System) 
    {
        Printf("Error: FMOD is not initialised.\n");
        return;
    }

    result = FMOD_System_GetNumDrivers(System, &num);
    result = FMOD_System_GetDriver(System, &current);

    for (i = 0; i < num; i++)
    {
        char name[512];
        int systemrate;
        FMOD_SPEAKERMODE speakermode;
        int speakermodechannels;

        result = FMOD_System_GetDriverInfo(System, i, name, sizeof(name), NULL, &systemrate, &speakermode, &speakermodechannels);

        Printf("%c%d. %s (%d Hz, %d channels)\n", (i == current ? '*' : ' '), i, name, systemrate, speakermodechannels);
    }
}

static void GetFmodFileUsage(long long *samp, long long *stre, long long *other)
{
    long long sampleBytesRead, streamBytesRead, otherBytesRead;
    FMOD_System_GetFileUsage(System, &sampleBytesRead, &streamBytesRead, &otherBytesRead);
    
    sampleBytesRead += FmodStats.SampleBytes;
    streamBytesRead += FmodStats.StreamBytes;
    
    *samp = sampleBytesRead;
    *stre = streamBytesRead;
    *other = otherBytesRead;
}

void FmodStatusCmd(int argc, char **argv)
{
    if (!System)
    {
        Printf("FMOD is not initialised.\n");
        return;
    }
    
    {
        int version, major, minor, patch;
        FMOD_System_GetVersion(System, &version);
        
        major = version >> 16;
        minor = (version & 0xFF00) >> 8;
        patch = version & 0xFF;
        
        Printf("FMOD Studio version %01x.%02x.%02x\n", major, minor, patch);
    }
    
    {
        int channels, realChannels, virtualChannels;
        FMOD_System_GetChannelsPlaying(System, &channels, &realChannels);
        virtualChannels = channels - realChannels;
        Printf("%d channels playing (%d real, %d virtual)\n", channels, realChannels, virtualChannels);
    }
    
    {
        float dsp, stream, geometry, update, total;
        FMOD_System_GetCPUUsage(System, &dsp, &stream, &geometry, &update, &total);
        Printf("CPU: DSP %0.2f%%; Stream %0.2f%%; Geometry %0.2f%%; Update %0.2f%%; Total %0.2f%%\n", dsp, stream, geometry, update, total);
    }
    
    {
        long long sampleBytesRead, streamBytesRead, otherBytesRead;
        GetFmodFileUsage(&sampleBytesRead, &streamBytesRead, &otherBytesRead);
        Printf("FSR: Sample %lld; Stream %lld; Other %lld\n", sampleBytesRead, streamBytesRead, otherBytesRead);
    }
}

void I_DrawSoundOverlay()
{
    int y = ST_Y - 16;
    char buf[512];
    
    if (!snd_drawoverlay->value)
    {
        return;
    }
    
    if (!System)
    {
        Con_DrawString(0, y, "FMOD is not initialised.");
        return;
    }

    {
        long long sampleBytesRead, streamBytesRead, otherBytesRead;
        GetFmodFileUsage(&sampleBytesRead, &streamBytesRead, &otherBytesRead);
        sprintf(buf, "FSR: Sample %lld; Stream %lld; Other %lld\n", sampleBytesRead, streamBytesRead, otherBytesRead);
        Con_DrawString(0, y, buf); y -= 16;
    }
    
    {
        float dsp, stream, geometry, update, total;
        FMOD_System_GetCPUUsage(System, &dsp, &stream, &geometry, &update, &total);
        sprintf(buf, "CPU: DSP %0.2f%%; Stream %0.2f%%; Geometry %0.2f%%; Update %0.2f%%; Total %0.2f%%\n", dsp, stream, geometry, update, total);
        Con_DrawString(0, y, buf); y -= 16;
    }
    
    {
        int channels, realChannels, virtualChannels;
        FMOD_System_GetChannelsPlaying(System, &channels, &realChannels);
        virtualChannels = channels - realChannels;
        sprintf(buf, "%d channels playing (%d real, %d virtual)\n", channels, realChannels, virtualChannels);
        Con_DrawString(0, y, buf); y -= 16;
    }
        
    {
        int version, major, minor, patch;
        FMOD_System_GetVersion(System, &version);
        
        major = version >> 16;
        minor = (version & 0xFF00) >> 8;
        patch = version & 0xFF;
    
        sprintf(buf, "FMOD Studio version %01x.%02x.%02x\n", major, minor, patch);
        Con_DrawString(0, y, buf); y -= 16;
    }
}

void I_InitSound()
{   
    FMOD_RESULT result;

    if (M_CheckParm("-nosound"))
    {
        nosound = true;
        return;
    }

    Printf("I_InitSound: Initialising FMOD Sound System.\n");

    result = FMOD_System_Create(&System);
    if (result != FMOD_OK)
    {
        I_CheckFmodError(result);
        Printf("Could not initialise FMOD - sound will be disabled.\n");
        return;
    }

    Printf("I_InitSound: FMOD Studio (c) Firelight Technologies Pty Ltd.\n");

    result = FMOD_System_Init(System, (int)snd_numchannels->value + 1, FMOD_INIT_NORMAL, NULL);
    if (result != FMOD_OK)
    {
        I_CheckFmodError(result);
        Printf("Could not initialise FMOD - sound will be disabled.\n");
        return;
    }

    result = FMOD_System_SetDriver(System, (int)snd_device->value);
    if (result != FMOD_OK)
    {
        I_CheckFmodError(result);
        Printf("Could not initialise FMOD (failed to set device) - sound will be disabled.\n");
        return;
    }

    result = FMOD_System_CreateChannelGroup(System, "Sounds", &SoundsChannels);
    if (result != FMOD_OK)
    {
        I_CheckFmodError(result);
        Printf("Could not initialise FMOD - sound will be disabled.\n");
        return;
    }

    result = FMOD_System_CreateChannelGroup(System, "Music", &MusicChannels);
    if (result != FMOD_OK)
    {
        I_CheckFmodError(result);
        Printf("Could not initialise FMOD - sound will be disabled.\n");
        return;
    }

    FMOD_System_Set3DSettings(System, 0.5f, 96.f, 1.f);

    nosound = false;
}

void I_ShutdownSound(void)
{
    FMOD_RESULT result;

    if (System)
        result = FMOD_System_Release(System);

    // Cleanup our globals
    // these are all deallocated by System::release anyway
    System = NULL;
    SoundsChannels = NULL;
    MusicChannels = NULL;
    MusicChannel = NULL;
    ActiveMusic = NULL;
}

// ... update sound buffer and audio device at runtime...
void I_UpdateSound(void)
{
    player_t *player;
    
    if (nosound) return;

    // update the listener
    player = &players[consoleplayer];
    if (player->mo)
    {
        mobj_t *mo = player->mo;

        // Note that y and z are swapped here - Doom is z=up, while FMOD is y=up
        FMOD_VECTOR pos = { FIXED2FLOAT(mo->x), FIXED2FLOAT(mo->z), FIXED2FLOAT(mo->y) };
        FMOD_VECTOR vel = { 0, 0, 0 };
        FMOD_VECTOR fwd = { FIXED2FLOAT(finecosine[mo->angle >> ANGLETOFINESHIFT]), 0, FIXED2FLOAT(finesine[mo->angle >> ANGLETOFINESHIFT]) };
        FMOD_VECTOR up = { 0, 1, 0 };
        FMOD_System_Set3DListenerAttributes(System, 0, &pos, &vel, &fwd, &up);
    }

    FMOD_System_Update(System);
}

// Get raw data lump index for sound descriptor.
int I_GetSfxLumpNum(sfxinfo_t *sfx)
{
    char namebuf[9];
    sprintf(namebuf, "ds%s", sfx->name);
    return W_GetNumForName(namebuf);
}

void SndReinitCmd(int argc, char **argv)
{
    I_ShutdownSound();
    I_InitSound();
}

///////////////////////////////////////////////////////////////////////////////
// Sound effects
///////////////////////////////////////////////////////////////////////////////
typedef struct
{
    FMOD_SOUND *Sound;
    FMOD_CHANNEL *Channel;
} FmodSoundEffectHandle;

// Starts a sound in a particular sound channel.
void *I_StartSound(int id, int vol, int sep, int pitch, int priority, mobj_t *source)
{
    char name[9];
    FMOD_SOUND *snd;
    FMOD_CHANNEL *ch;
    FmodSoundEffectHandle *handle;

    if (nosound) return NULL;

    sprintf(name, "DS%s", S_sfx[id].name);

    snd = LoadFmodSound(name, &ch, (source ? FMOD_3D : FMOD_2D));

    handle = (FmodSoundEffectHandle *)M_Malloc(sizeof(FmodSoundEffectHandle));
    handle->Sound = snd;
    handle->Channel = ch;

    FMOD_Channel_SetVolume(ch, (float)vol / 127.0f);
    FMOD_Channel_SetPitch(ch, (float)pitch / 127.0f);

    if (source)
    {
        // Note that y and z are swapped here - Doom is z=up, while FMOD is y=up
        FMOD_VECTOR pos = {FIXED2FLOAT(source->x), FIXED2FLOAT(source->z), FIXED2FLOAT(source->y)};
        FMOD_VECTOR vel = {0, 0, 0};
        FMOD_Channel_Set3DAttributes(ch, &pos, &vel);
    }

    return handle;
}

// Stops a sound channel.
void I_StopSound(void *handle)
{
    FmodSoundEffectHandle *hnd = (FmodSoundEffectHandle *)handle;

    if (nosound) return;

    FMOD_Channel_Stop(hnd->Channel);
    M_Free(hnd);
}

// Called by S_*() functions
//  to see if a channel is still playing.
// Returns 0 if no longer playing, 1 if playing.
int I_SoundIsPlaying(void *handle)
{
    FmodSoundEffectHandle *hnd = (FmodSoundEffectHandle *)handle;
    FMOD_BOOL isPlaying;

    if (nosound) return false;

    FMOD_Channel_IsPlaying(hnd->Channel, &isPlaying);

    return isPlaying;
}

// Updates the volume, separation,
//  and pitch of a sound channel.
void I_UpdateSoundParams(void *handle, mobj_t *source, int vol, int sep, int pitch)
{
    FmodSoundEffectHandle *hnd = (FmodSoundEffectHandle *)handle;

    if (nosound) return;

    FMOD_Channel_SetVolume(hnd->Channel, (float)vol / 127.0f);
    FMOD_Channel_SetPitch(hnd->Channel, (float)pitch / 127.0f);

    if (source)
    {
        // Note that y and z are swapped here - Doom is z=up, while FMOD is y=up
        FMOD_VECTOR pos = {FIXED2FLOAT(source->x), FIXED2FLOAT(source->z), FIXED2FLOAT(source->y)};
        FMOD_VECTOR vel = {0, 0, 0};
        FMOD_Channel_Set3DAttributes(hnd->Channel, &pos, &vel);
    }
}

// Volume.
void I_SetMusicVolume(int volume)
{
    if (nosound) return;

    FMOD_ChannelGroup_SetVolume(MusicChannels, (float)volume / 127.0f);
}

///////////////////////////////////////////////////////////////////////////////
// Game music
///////////////////////////////////////////////////////////////////////////////
void I_PauseSong()
{
    if (!MusicChannel) return;
    if (nosound) return;

    FMOD_Channel_SetPaused(MusicChannel, true);
}

void I_ResumeSong()
{
    if (!MusicChannel) return;
    if (nosound) return;

    FMOD_Channel_SetPaused(MusicChannel, false);
}


byte *ConvertMusToMidi(int lump, unsigned *outSize)
{
    int size;
    FILE *musf, *midf;
    byte *musbuf, *midbuf;
    int ow = 2;
    int r;

    size = W_LumpLength(lump);
    musbuf = (byte *)M_Malloc(size);

    W_ReadLump(lump, musbuf);

    // Write this to a temporary file...
    musf = fopen("doommus.tmp", "wb");
    if (!musf)
    {
        Printf("Failed to convert %s to MIDI.\n", lumpinfo[lump].name);
        return NULL;
    }

    fwrite(musbuf, sizeof(byte), size, musf);
    fclose(musf);
    
    M_Free(musbuf);

    convert("doommus.tmp", "doommid.tmp", 1, 89, 0, 1, &ow);

    midf = fopen("doommid.tmp", "rb");
    if (!midf)
    {
        Printf("Failed to convert %s to MIDI.\n", lumpinfo[lump].name);
        return NULL;
    }

    fseek(midf, 0, SEEK_END);
    size = ftell(midf);
    fseek(midf, 0, SEEK_SET);

    midbuf = (byte *)M_Malloc(size);
    r = fread(midbuf, sizeof(byte), size, midf);
    fclose(midf);
    if (r < size) 
    {
        Printf("Failed to load MUS file\n");
        return NULL;
    }

    *outSize = size;

    remove("doommus.tmp");
    remove("doommid.tmp");

    return midbuf;
}

// Registers a song handle to song data.
void I_RegisterSong(const char *name)
{
    FMOD_RESULT result;
    FMOD_SOUND *snd;
    FMOD_CREATESOUNDEXINFO exinfo;
    FMOD_MODE mode;
    int lump;
    byte *header;
    boolean ismus = false;
    int length;
    const char *fn = name;
    FmodUserData *userdata;

    if (nosound) return;

    lump = W_CheckNumForName((char *)name);

    if (lump == -1)
        return;

    memset(&exinfo, 0, sizeof(FMOD_CREATESOUNDEXINFO));
    exinfo.cbsize = sizeof(FMOD_CREATESOUNDEXINFO);

    mode = FMOD_2D|FMOD_CREATESTREAM|FMOD_LOOP_NORMAL;

    // Check if this song is a MUS file, and if so, convert it to MIDI
    {
        unsigned c;

        header = (byte *)alloca(3);
        c = W_ReadPartialLump(lump, header, 0, 3);

        if (c >= 3 && memcmp(header, "MUS", 3) == 0)
        {
            ismus = true;

            fn = (const char *)ConvertMusToMidi(lump, (unsigned int *)&length);
            if (fn == NULL)
            {
                return;
            }

            mode |= FMOD_OPENMEMORY;
            mode &= ~FMOD_CREATESTREAM; // can't stream it if it's already all in RAM!
            exinfo.length = length;
        }
    }

    if (!ismus)
        AddFmodCallbacks(&exinfo);
        
    userdata = M_Malloc(sizeof(FmodUserData));
    memset(userdata, 0, sizeof(FmodUserData));
    userdata->IsStream = (mode & FMOD_CREATESTREAM);
    exinfo.fileuserdata = userdata;

    result = FMOD_System_CreateSound(System, fn, mode, &exinfo, &snd);
    if (result != FMOD_OK)
    {
        Printf("Failed to play song %s\n", name);
    }
    
    if (ismus) M_Free((void *)fn);
    ActiveMusic = snd;
}

// Called by anything that wishes to start music.
//  plays a song, and when the song is done,
//  starts playing it again in an endless loop.
// Horrible thing to do, considering.
void I_PlaySong(int looping)
{
    if (nosound) return;

    FMOD_Sound_SetLoopCount(ActiveMusic, (looping ? -1 : 0));
    FMOD_System_PlaySound(System, ActiveMusic, MusicChannels, false, &MusicChannel);
    FMOD_ChannelGroup_SetVolume(MusicChannels, snd_musvolume->value / 127.0f);
}

// Stops a song over 3 seconds.
void I_StopSong()
{
    if (!MusicChannel) return;
    if (nosound) return;

    FMOD_Channel_Stop(MusicChannel);
}

// See above (register), then think backwards
void I_UnRegisterSong()
{
    if (!ActiveMusic) return;
    if (nosound) return;

    FMOD_Sound_Release(ActiveMusic);
    ActiveMusic = NULL;
    MusicChannel = NULL;
}

FMOD_SOUND *LoadFmodSound(const char *name, FMOD_CHANNEL **chan, FMOD_MODE pmode)
{
    byte *data;
    int lump, size;
    FMOD_RESULT result;
    FMOD_SOUND *snd;
    FMOD_CREATESOUNDEXINFO exinfo;
    stream_t headStream;
    int sampleRate, numSamples;
    byte *samplesStart;
    FMOD_MODE mode;
    FmodUserData *userdata;

    memset(&exinfo, 0, sizeof(FMOD_CREATESOUNDEXINFO));
    exinfo.cbsize = sizeof(FMOD_CREATESOUNDEXINFO);

    mode = pmode|FMOD_OPENMEMORY;

    lump = W_GetNumForName((char *)name);
    size = W_LumpLength(lump);
    data = (byte *)M_Malloc(size);

    W_ReadLump(lump, data);
    FmodStats.SampleBytes += size;

    Stream_InitSize(&headStream, data, size, false);

    DevPrintf("Loading in sound lump %s\n", name);

    // Is this a Doom DMX sound?
    if (Stream_ReadByte(&headStream) == 0x03 && Stream_ReadByte(&headStream) == 0x00)
    {
        sampleRate = Stream_ReadShort(&headStream);
        numSamples = Stream_ReadInt(&headStream);

        if (numSamples <= 0)
        {
            M_Free(data);
            Printf("sound %s has no samples! (%d)\n", name, numSamples);
            return NULL;
        }

        samplesStart = headStream.orig + 8;
        
        // The sample data is sandwiched between 16 bytes of padding on each
        // side
        if (snd_ignorepadding->value)
        {
            samplesStart += 16;
            numSamples -= 32;
        }

        mode |= FMOD_OPENRAW;

        exinfo.length = numSamples;
        exinfo.numchannels = 1;
        exinfo.defaultfrequency = sampleRate;
        exinfo.format = FMOD_SOUND_FORMAT_PCM8;

        DevPrintf("DMX format data; %d samples at %d Hz\n", numSamples, sampleRate);
    }
    // Just let FMOD figure it out
    else
    {
        samplesStart = headStream.orig;
        exinfo.length = size;
        DevPrintf("Sound is of unknown format; going to let FMOD figure it out\n");
    }

    userdata = M_Malloc(sizeof(FmodUserData));
    memset(userdata, 0, sizeof(FmodUserData));
    userdata->IsStream = (mode & FMOD_CREATESTREAM);
    exinfo.fileuserdata = userdata;
    
    result = FMOD_System_CreateSound(System, (const char *)samplesStart, mode, &exinfo, &snd);
    if (result != FMOD_OK)
    {
        Printf("Failed to create sound %s\n", name);
    }
    
    M_Free(data);

    result = FMOD_System_PlaySound(System, snd, SoundsChannels, 0, chan);
    if (result != FMOD_OK)
    {
        Printf("Failed to play sound %s\n", name);
    }

    return snd;
}

///////////////////////////////////////////////////////////////////////////////
// FMOD file callbacks
///////////////////////////////////////////////////////////////////////////////
typedef struct {
    char Name[9];
    int Lump;
    int Size;
    unsigned Pos;
} FmodSoundHandle;

FMOD_RESULT F_CALLBACK FmodOpenFile(const char *name, unsigned *filesize, void **handle, void *userdata)
{
    int lump, size;
    FmodSoundHandle *fhandle;
    FmodUserData *user = (FmodUserData *)userdata;

    lump = W_CheckNumForName((char *)name);
    if (lump == -1)
        return FMOD_ERR_FILE_NOTFOUND;

    size = W_LumpLength(lump);

    fhandle = (FmodSoundHandle *)M_Malloc(sizeof(FmodSoundHandle));
    memset(fhandle, 0, sizeof(FmodSoundHandle));

    fhandle->Lump = lump;
    fhandle->Size = size;
    fhandle->Pos = 0;

    memset(fhandle->Name, 0, 9);
    strcpy(fhandle->Name, name);

    *handle = fhandle;
    *filesize = size;

    return FMOD_OK;
}

FMOD_RESULT F_CALLBACK FmodCloseFile(void *handle, void *userdata)
{
    FmodSoundHandle *fhandle = (FmodSoundHandle *)handle;
    FmodUserData *user = (FmodUserData *)userdata;

    M_Free(fhandle);
    
    if (userdata)
    {
        M_Free(userdata);
    }

    return FMOD_OK;
}

FMOD_RESULT F_CALLBACK FmodReadFile(void *handle, void *buffer, unsigned sizebytes, unsigned *bytesread, void *userdata)
{
    FmodSoundHandle *fhandle = (FmodSoundHandle *)handle;
    FmodUserData *user = (FmodUserData *)userdata;
    unsigned c;

    c = W_ReadPartialLump(fhandle->Lump, buffer, fhandle->Pos, sizebytes);
    fhandle->Pos += c;

    *bytesread = c;
    if (user->IsStream)
        FmodStats.StreamBytes += c;
    else
        FmodStats.SampleBytes += c;

    return FMOD_OK;
}

FMOD_RESULT F_CALLBACK FmodSeekFile(void *handle, unsigned pos, void *userdata)
{
    FmodSoundHandle *fhandle = (FmodSoundHandle *)handle;
    FmodUserData *user = (FmodUserData *)userdata;

    if (pos < 0 || (int)pos >= fhandle->Size)
        return FMOD_ERR_INVALID_POSITION;

    fhandle->Pos = pos;

    return FMOD_OK;
}

void AddFmodCallbacks(FMOD_CREATESOUNDEXINFO *exinfo)
{
    exinfo->fileuseropen = FmodOpenFile;
    exinfo->fileuserclose = FmodCloseFile;
    exinfo->fileuserread = FmodReadFile;
    exinfo->fileuserseek = FmodSeekFile;
}
