//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// Copyright 1993-1997 id Software
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// DESCRIPTION:
//	Memory allocation wrappers
//
//-----------------------------------------------------------------------------

#include <stdlib.h>
#include "m_alloc.h"
#include "i_system.h"
#include "c_console.h"

#if defined(_MSC_VER) && defined(_DEBUG)
#define M_MALLOC_USE_DBG
#include <crtdbg.h>
#endif

void *_M_Malloc(size_t size, const char *file, int line, const char *function)
{
    void *block;
    
#ifdef M_MALLOC_USE_DBG
    block = _malloc_dbg(size, _NORMAL_BLOCK, file, line);
#else
    block = malloc(size);
#endif

    if (block == NULL)
    {
        I_Error("Failed to malloc %u bytes\n\tat %s %s line %d", size, function, file, line);
    }

    return block;
}

void *_M_Calloc(size_t num, size_t size, const char *file, int line, const char *function)
{
    void *block;
    
#ifdef M_MALLOC_USE_DBG
    block = _calloc_dbg(num, size, _NORMAL_BLOCK, file, line);
#else
    block = calloc(num, size);
#endif

    if (block == NULL)
    {
        I_Error("Failed to calloc %u blocks of size %u (= %u total bytes)\n\tat %s %s line %d", num, size, num * size, function, file, line);
    }

    return block;
}

void *_M_Realloc(void *block, size_t newsize, const char *file, int line, const char *function)
{
    void *newblock;
    
#ifdef M_MALLOC_USE_DBG
    newblock = _realloc_dbg(block, newsize, _NORMAL_BLOCK, file, line);
#else
    newblock = realloc(block, newsize);
#endif

    if (newblock == NULL)
    {
        I_Error("Failed to realloc %p to %u bytes\n\tat %s %s line %d", block, newsize, function, file, line);
    }

    return newblock;
}

void _M_Free(void *block, const char *file, int line, const char *function)
{
#ifdef _DEBUG
    if (block == NULL)
    {
        Printf("WARNING! Attempted to M_Free invalid pointer %p\n\tat %s %s line %d\n", block, function, file, line);

        return;
    }
#endif

#ifdef M_MALLOC_USE_DBG
    _free_dbg(block, _NORMAL_BLOCK);
#else
    free(block);
#endif
}