//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// Copyright 1993-1997 id Software
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// DESCRIPTION:
//	Main loop menu stuff.
//	Default Config File.
//
//-----------------------------------------------------------------------------

#include <sys/types.h>
#include <fcntl.h>
#include <stdlib.h>

#include <ctype.h>

#include "doomdef.h"

#include "z_zone.h"

#include "m_swap.h"
#include "m_argv.h"

#include "w_wad.h"

#include "i_system.h"
#include "i_video.h"
#include "multires.h"

#include "hu_stuff.h"

// State.
#include "doomstat.h"

// Data.
#include "dstrings.h"

#include "m_misc.h"
#include "c_binds.h"
#include "c_cvar.h"
#include "c_console.h"
#include "c_cmds.h"
#include "m_alloc.h"
#include "version.h"
#include "hginfo.h"

#ifdef _MSC_VER
#include <io.h>
#endif

//
// M_DrawText
// Returns the final X coordinate
// HU_Init must have been called to init the font
//
extern patch_t *hu_font[HU_FONTSIZE];
extern double gammalevel;

int M_DrawText(int x,
               int y,
               boolean direct,
               char *string)
{
    int c;
    int w;

    while (*string)
    {
        c = toupper(*string) - HU_FONTSTART;
        string++;
        if (c < 0 || c > HU_FONTSIZE)
        {
            x += 4;
            continue;
        }

        w = SHORT(hu_font[c]->width);
        if (x + w > SCREENWIDTH)
            break;
        if (direct)
            V_DrawPatchDirect(x, y, &screens[0], hu_font[c]);
        else
            V_DrawPatch(x, y, &screens[0], hu_font[c]);
        x += w;
    }

    return x;
}

//
// M_WriteFile
//
#ifndef O_BINARY
#define O_BINARY 0
#endif

boolean
M_WriteFile(char const *name,
            void *source,
            int length)
{
    FILE *handle;
    int count;

    handle = fopen(name, "wb");

    if (handle == NULL)
        return false;

    count = fwrite(source, sizeof(byte), length, handle);
    fclose(handle);

    if (count < length)
        return false;

    return true;
}

//
// M_ReadFile
//
int M_ReadFile(char const *name,
               byte **buffer)
{
    FILE *handle;
    int count, length;
    byte *buf;

    handle = fopen(name, "rb");
    if (handle == NULL)
        I_Error("Couldn't read file %s", name);
        
    fseek(handle, 0, SEEK_END);
    length = ftell(handle);
    fseek(handle, 0, SEEK_SET);
    
    buf = (byte *)Z_Malloc(length, PU_STATIC, NULL);
    count = fread(buf, sizeof(byte), length, handle);
    fclose(handle);

    if (count < length)
        I_Error("Couldn't read file %s", name);

    *buffer = buf;
    return length;
}

//
// M_SaveDefaults
//
void M_SaveDefaults(void)
{
    FILE *f;
    char filename[256];
    char header[256];
    char gammacmd[256];

    memset(filename, 0, 256);
    sprintf(filename, "%s/%s.cfg", I_GetConfigDir(), PACKAGE_ID);

    memset(header, 0, 256);
    sprintf(header, "// Automatically generated by %s v%s (r%s)\n", PACKAGE_NAME, PACKAGE_VERSION, HG_REVISION_STR);
    sprintf(header + strlen(header), "// Edit at your own risk! This file is overwritten every time you exit %s.\n", PACKAGE_NAME);

    memset(header, 0, 256);
    sprintf(gammacmd, "gamma \"%f\"\n", gammalevel);

    f = fopen(filename, "w");
    fwrite(header, sizeof(char), strlen(header), f);

    C_SaveCvars(f);
    C_SaveBindings(f);
    fwrite(gammacmd, sizeof(char), strlen(gammacmd), f);

    fclose(f);
}

//
// M_LoadDefaults
//
extern byte scantokey[128];

void M_LoadDefaults(void)
{
    char filename[256];
    char *execcmd;
    FILE *f;

    memset(filename, 0, 256);
    sprintf(filename, "%s/%s.cfg", I_GetConfigDir(), PACKAGE_ID);

    // load defaults if config doesn't exist
    f = fopen(filename, "r");
    if (!f)
    {
        int i;
        for (i = 0; i < 10; i++)
        {
            char kn[2];
            char cmd[32];
            kn[0] = '0' + i;
            kn[1] = 0;
            cmd[0] = 0;
            sprintf(cmd, "slot %d", i);
            C_RegisterBinding(kn, cmd);
        }

        C_RegisterBinding("`", "toggleconsole");

        C_ExecuteString("gamma 1");

        return;
    }

    fclose(f);

    // Load in CVars
    execcmd = (char *)M_Malloc(strlen(filename) + 8);
    sprintf(execcmd, "exec \"%s\"", filename);
    C_ExecuteMultiple(execcmd);
    M_Free(execcmd);
}
