//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// Copyright 1993-1997 id Software
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// DESCRIPTION:
//
//
//-----------------------------------------------------------------------------

#ifndef __M_RANDOM__
#define __M_RANDOM__

#include "doomtype.h"

// killough 1/19/98: rewritten to use to use a better random number generator
// in the new engine, although the old one is available for compatibility.

// killough 2/16/98:
//
// Make every random number generator local to each control-equivalent block.
// Critical for demo sync. Changing the order of this list breaks all previous
// versions' demos. The random number generators are made local to reduce the
// chances of sync problems. In Doom, if a single random number generator call
// was off, it would mess up all random number generators. This reduces the
// chances of it happening by making each RNG local to a control flow block.
//
// Notes to developers: if you want to reduce your demo sync hassles, follow
// this rule: for each call to P_Random you add, add a new class to the enum
// type below for each block of code which calls P_Random. If two calls to
// P_Random are not in "control-equivalent blocks", i.e. there are any cases
// where one is executed, and the other is not, put them in separate classes.
//
// Keep all current entries in this list the same, and in the order
// indicated by the #'s, because they're critical for preserving demo
// sync. Do not remove entries simply because they become unused later.

typedef enum {
	pr_skullfly,                // #1
	pr_damage,                  // #2
	pr_crush,                   // #3
	pr_genlift,                 // #4
	pr_killtics,                // #5
	pr_damagemobj,              // #6
	pr_painchance,              // #7
	pr_lights,                  // #8
	pr_explode,                 // #9
	pr_respawn,                 // #10
	pr_lastlook,                // #11
	pr_spawnthing,              // #12
	pr_spawnpuff,               // #13
	pr_spawnblood,              // #14
	pr_missile,                 // #15
	pr_shadow,                  // #16
	pr_plats,                   // #17
	pr_punch,                   // #18
	pr_punchangle,              // #19
	pr_saw,                     // #20
	pr_plasma,                  // #21
	pr_gunshot,                 // #22
	pr_misfire,                 // #23
	pr_shotgun,                 // #24
	pr_bfg,                     // #25
	pr_slimehurt,               // #26
	pr_dmspawn,                 // #27
	pr_missrange,               // #28
	pr_trywalk,                 // #29
	pr_newchase,                // #30
	pr_newchasedir,             // #31
	pr_see,                     // #32
	pr_facetarget,              // #33
	pr_posattack,               // #34
	pr_sposattack,              // #35
	pr_cposattack,              // #36
	pr_spidrefire,              // #37
	pr_troopattack,             // #38
	pr_sargattack,              // #39
	pr_headattack,              // #40
	pr_bruisattack,             // #41
	pr_tracer,                  // #42
	pr_skelfist,                // #43
	pr_scream,                  // #44
	pr_brainscream,             // #45
	pr_cposrefire,              // #46
	pr_brainexp,                // #47
	pr_spawnfly,                // #48
	pr_misc,                    // #49
	pr_all_in_one,              // #50
	// Start new entries -- add new entries below

	// End of new entries
	NUMPRCLASS               // MUST be last item in list
} pr_class_t;

// The random number generator's state.
typedef struct {
	unsigned long seed[NUMPRCLASS];      // Each block's random seed
	int rndindex, prndindex;             // For compatibility support
} rng_t;

extern rng_t rng;                      // The rng's state

extern unsigned long rngseed;          // The starting seed (not part of state)

// Returns a number from 0 to 255,
#define M_Random() P_Random(pr_misc)

// As M_Random, but used by the play simulation.
int P_Random(pr_class_t);

// Fix randoms for demos.
void M_ClearRandom(void);

#endif
