//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// Copyright 1993-1997 id Software
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
//-----------------------------------------------------------------------------

#include <string.h>
#include "m_stream.h"
#include "m_alloc.h"
#include "i_system.h"

#define MAX_STREAM_STRING_SIZE 2048

static boolean Stream_IsSpaceAvailable(stream_t *stream, size_t bytes);

void Stream_Alloc(stream_t *stream, size_t length, boolean dynamic)
{
    byte *block = (byte *)M_Malloc(length);

    stream->data = block;
    stream->dataEnd = block + length;
    stream->orig = block;
    stream->dynamic = dynamic;
    stream->managed = true;
}

void Stream_InitSize(stream_t *stream, byte *buffer, size_t size, boolean dynamic)
{
    stream->data = buffer;
    stream->dataEnd = buffer + size;
    stream->orig = buffer;
    stream->dynamic = dynamic;
    stream->managed = false;
}

void Stream_InitPtr(stream_t *stream, byte *begin, byte *end, boolean dynamic)
{
    stream->data = begin;
    stream->dataEnd = end;
    stream->orig = begin;
    stream->dynamic = dynamic;
    stream->managed = false;
}

void Stream_Free(stream_t *stream)
{
    if (stream->managed && stream->orig)
    {
        M_Free(stream->orig);
    }

    memset(stream, 0, sizeof(stream_t));
}

boolean Stream_IsSpaceAvailable(stream_t *stream, size_t bytes)
{
    boolean avail = !(stream->data + bytes > stream->dataEnd);

    if (stream->dynamic && !avail)
    {
        size_t additional = 512;
        size_t cursize = Stream_GetSize(stream);
        size_t newsize;
        byte *newbuf;

        while (additional < bytes)
            additional += 512;

        newsize = cursize + additional;
        newbuf = (byte *)M_Realloc(stream->orig, newsize);

        stream->data = newbuf + Stream_GetPosition(stream);
        stream->orig = newbuf;
        stream->dataEnd = newbuf + newsize;

        return true;
    }
    else
    {
        return avail;
    }
}

size_t Stream_GetSize(stream_t *stream)
{
    return stream->dataEnd - stream->orig;
}

size_t Stream_GetPosition(stream_t *stream)
{
    return stream->data - stream->orig;
}

size_t Stream_GetAvailable(stream_t *stream)
{
    return stream->dataEnd - stream->data;
}

boolean Stream_Write(stream_t *stream, byte *source, size_t length)
{
    if (!Stream_IsSpaceAvailable(stream, length))
        return false;

    memcpy(stream->data, source, length);
    stream->data += length;

    return true;
}

boolean Stream_WriteByte(stream_t *stream, byte value)
{
    if (!Stream_IsSpaceAvailable(stream, sizeof(byte)))
        return false;

    stream->data[0] = value;

    stream->data += sizeof(byte);

    return true;
}

boolean Stream_WriteShort(stream_t *stream, short value)
{
    if (!Stream_IsSpaceAvailable(stream, sizeof(short)))
        return false;

#ifndef __BIG_ENDIAN__
    stream->data[0] = value & 0xFF;
    stream->data[1] = (value >> 8) & 0xFF;
#else
    stream->data[1] = value & 0xFF;
    stream->data[0] = (value >> 8) & 0xFF;
#endif

    stream->data += sizeof(short);

    return true;
}

boolean Stream_WriteInt(stream_t *stream, int value)
{
    if (!Stream_IsSpaceAvailable(stream, sizeof(int)))
        return false;

#ifndef __BIG_ENDIAN__
    stream->data[0] = value & 0xFF;
    stream->data[1] = (value >> 8) & 0xFF;
    stream->data[2] = (value >> 16) & 0xFF;
    stream->data[3] = (value >> 24) & 0xFF;
#else
    stream->data[3] = value & 0xFF;
    stream->data[2] = (value >> 8) & 0xFF;
    stream->data[1] = (value >> 16) & 0xFF;
    stream->data[0] = (value >> 24) & 0xFF;
#endif

    stream->data += sizeof(int);

    return true;
}

boolean Stream_WriteString(stream_t *stream, const char *value)
{
    size_t len, size;

    len = strlen(value);
    size = len + 1;

    if (!Stream_IsSpaceAvailable(stream, size))
        return false;

    Stream_Write(stream, (byte *)value, len);
    Stream_WriteByte(stream, 0x00);

    return true;
}

boolean Stream_Read(stream_t *stream, byte *dest, size_t length)
{
    if (!Stream_IsSpaceAvailable(stream, length))
        return false;

    memcpy(dest, stream->data, length);
    stream->data += length;

    return true;
}

byte Stream_ReadByte(stream_t *stream)
{
    byte value;

    if (!Stream_IsSpaceAvailable(stream, sizeof(byte)))
        I_Error("Attempt to read past stream\n");

    value = stream->data[0];
    stream->data += sizeof(byte);

    return value;
}

short Stream_ReadShort(stream_t *stream)
{
    short value;

    if (!Stream_IsSpaceAvailable(stream, 2))
        I_Error("Attempt to read past stream\n");

#ifndef __BIG_ENDIAN__
    value = stream->data[0] + (stream->data[1] << 8);
#else
    value = stream->data[1] + (stream->data[0] << 8);
#endif

    stream->data += 2;

    return value;
}

int Stream_ReadInt(stream_t *stream)
{
    int value;

    if (!Stream_IsSpaceAvailable(stream, 4))
        I_Error("Attempt to read past stream\n");

#ifndef __BIG_ENDIAN__
    value = stream->data[0] + (stream->data[1] << 8) + (stream->data[2] << 16) + (stream->data[3] << 24);
#else
    value = stream->data[3] + (stream->data[2] << 8) + (stream->data[1] << 16) + (stream->data[0] << 24);
#endif

    stream->data += 4;

    return value;
}

const char *Stream_ReadString(stream_t *stream)
{
    static char value[MAX_STREAM_STRING_SIZE];
    size_t i = 0, len = 0, end;

    memset(value, 0, MAX_STREAM_STRING_SIZE);

    do
    {
        char c = (char)Stream_ReadByte(stream);

        if (c == 0)
            break;

        if (i < MAX_STREAM_STRING_SIZE - 1)
            value[i] = c;

        i++;
    } while (true);

    end = i < MAX_STREAM_STRING_SIZE ? i : MAX_STREAM_STRING_SIZE - 1;
    value[end] = '\0';

    return value;
}