//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// Copyright 1993-1997 id Software
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
//
// 3. Neither the name of the copyright holder nor the names of its
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// DESCRIPTION:
//  Archiving: SaveGame I/O.
//
//-----------------------------------------------------------------------------

#include "p_saveg.h"
#include "i_system.h"
#include "z_zone.h"
#include "p_local.h"
#include "doomstat.h"
#include "r_state.h"
#include "m_alloc.h"
#include "p_map.h"

stream_t SaveStream;

//
// P_ArchivePlayers
//
void P_ArchivePlayers(void)
{
    int i;
    int j;
    player_t *dest;

    for (i = 0; i < MAXPLAYERS; i++)
    {
        if (!playeringame[i])
            continue;

        dest = (player_t *)SaveStream.data;
        memcpy(dest, &players[i], sizeof(player_t));
        SaveStream.data += sizeof(player_t);
        for (j = 0; j < NUMPSPRITES; j++)
        {
            if (dest->psprites[j].state)
            {
                dest->psprites[j].state = (state_t *)(dest->psprites[j].state - states);
            }
        }
    }
}

//
// P_UnArchivePlayers
//
void P_UnArchivePlayers(void)
{
    int i;
    int j;

    for (i = 0; i < MAXPLAYERS; i++)
    {
        if (!playeringame[i])
            continue;

        memcpy(&players[i], SaveStream.data, sizeof(player_t));
        SaveStream.data += sizeof(player_t);

        // will be set when unarc thinker
        players[i].mo = NULL;
        players[i].message = NULL;
        players[i].attacker = NULL;

        for (j = 0; j < NUMPSPRITES; j++)
        {
            if (players[i].psprites[j].state)
            {
                players[i].psprites[j].state = &states[(dintptr)players[i].psprites[j].state];
            }
        }
    }
}

//
// P_ArchiveWorld
//
void P_ArchiveWorld(void)
{
    int i;
    int j;
    sector_t *sec;
    line_t *li;
    side_t *si;

    // do sectors
    for (i = 0, sec = sectors; i < numsectors; i++, sec++)
    {
        Stream_WriteInt(&SaveStream, sec->floorheight);
        Stream_WriteInt(&SaveStream, sec->ceilingheight);
        Stream_WriteShort(&SaveStream, sec->floorpic);
        Stream_WriteShort(&SaveStream, sec->ceilingpic);
        Stream_WriteShort(&SaveStream, sec->lightlevel);
        Stream_WriteShort(&SaveStream, sec->special);
        Stream_WriteShort(&SaveStream, sec->tag);
    }

    // do lines
    for (i = 0, li = lines; i < numlines; i++, li++)
    {
        Stream_WriteShort(&SaveStream, li->flags);
        Stream_WriteShort(&SaveStream, li->special);
        Stream_WriteShort(&SaveStream, li->tag);
        for (j = 0; j < 2; j++)
        {
            if (li->sidenum[j] == -1)
                continue;

            si = &sides[li->sidenum[j]];

            Stream_WriteInt(&SaveStream, si->textureoffset);
            Stream_WriteInt(&SaveStream, si->rowoffset);
            Stream_WriteShort(&SaveStream, si->toptexture);
            Stream_WriteShort(&SaveStream, si->bottomtexture);
            Stream_WriteShort(&SaveStream, si->midtexture);
        }
    }
}

//
// P_UnArchiveWorld
//
void P_UnArchiveWorld(void)
{
    int i;
    int j;
    sector_t *sec;
    line_t *li;
    side_t *si;

    // do sectors
    for (i = 0, sec = sectors; i < numsectors; i++, sec++)
    {
        sec->floorheight = Stream_ReadInt(&SaveStream);
        sec->ceilingheight = Stream_ReadInt(&SaveStream);
        sec->floorpic = Stream_ReadShort(&SaveStream);
        sec->ceilingpic = Stream_ReadShort(&SaveStream);
        sec->lightlevel = Stream_ReadShort(&SaveStream);
        sec->special = Stream_ReadShort(&SaveStream);
        sec->tag = Stream_ReadShort(&SaveStream);
        sec->specialdata = 0;

        // phares 9/13/98: soundtarget has meaning, to save sound info across
        // savegames.
        sec->soundtarget = (mobj_t *) ((dintptr) Stream_ReadShort(&SaveStream));
    }

    // do lines
    for (i = 0, li = lines; i < numlines; i++, li++)
    {
        li->flags = Stream_ReadShort(&SaveStream);
        li->special = Stream_ReadShort(&SaveStream);
        li->tag = Stream_ReadShort(&SaveStream);
        for (j = 0; j < 2; j++)
        {
            if (li->sidenum[j] == -1)
                continue;
            si = &sides[li->sidenum[j]];
            si->textureoffset = Stream_ReadInt(&SaveStream);
            si->rowoffset = Stream_ReadInt(&SaveStream);
            si->toptexture = Stream_ReadShort(&SaveStream);
            si->bottomtexture = Stream_ReadShort(&SaveStream);
            si->midtexture = Stream_ReadShort(&SaveStream);
        }
    }
}

//
// Thinkers
//
typedef enum {
    tc_end,
    tc_mobj
} thinkerclass_t;

// phares 9/13/98: Moved this code outside of P_ArchiveThinkers so the
// thinker indices could be used by the code that saves sector info.
static int number_of_thinkers;

void P_ThinkerToIndex(void)
{
    thinker_t *th;

    // killough 2/14/98:
    // count the number of thinkers, and mark each one with its index, using
    // the prev field as a placeholder, since it can be restored later.
    number_of_thinkers = 0;
    for (th = thinkercap.next ; th != &thinkercap ; th=th->next)
        if (th->function.acp1 == (actionf_p1) P_MobjThinker)
            th->prev = (thinker_t *) ++number_of_thinkers;
}

// phares 9/13/98: Moved this code outside of P_ArchiveThinkers so the
// thinker indices could be used by the code that saves sector info.
void P_IndexToThinker(void)
{
    // killough 2/14/98: restore prev pointers
    thinker_t *th;
    thinker_t *prev = &thinkercap;

    for (th = thinkercap.next ; th != &thinkercap ; prev=th, th=th->next)
        th->prev = prev;
}


//
// P_ArchiveThinkers
//
void P_ArchiveThinkers(void)
{
    thinker_t *th;

    for (th = thinkercap.next ; th != &thinkercap ; th=th->next)
        if (th->function.acp1 == (actionf_p1) P_MobjThinker)
        {
            mobj_t *mobj;

            Stream_WriteByte(&SaveStream, tc_mobj);
            mobj = (mobj_t *)SaveStream.data;
            memcpy (mobj, th, sizeof(*mobj));
            SaveStream.data += sizeof(*mobj);
            mobj->state = (state_t *)(mobj->state - states);

            // killough 2/14/98: convert pointers into indices.
            // Fixes many savegame problems, by properly saving
            // target and tracer fields. Note: we store NULL if
            // the thinker pointed to by these fields is not a
            // mobj thinker.

            if (mobj->target)
                mobj->target = mobj->target->thinker.function.acp1 ==
                    (actionf_p1) P_MobjThinker ?
                    (mobj_t *) mobj->target->thinker.prev : NULL;

            if (mobj->tracer)
                mobj->tracer = mobj->tracer->thinker.function.acp1 ==
                    (actionf_p1) P_MobjThinker ?
                    (mobj_t *) mobj->tracer->thinker.prev : NULL;

            if (mobj->player)
                mobj->player = (player_t *)((mobj->player-players) + 1);
        }

    // add a terminating marker
    Stream_WriteByte(&SaveStream, tc_end);
}

//
// P_UnArchiveThinkers
//
void P_UnArchiveThinkers(void)
{
    thinker_t *th;
    mobj_t    **mobj_p;    // killough 2/14/98: Translation table
    size_t    size;        // killough 2/14/98: size of or index into table
    int       i;           // phares 9/13/98:   For sec->soundtarget restore
    sector_t* sec;         // phares 9/13/98:   For sec->soundtarget restore

    // remove all the current thinkers
    for (th = thinkercap.next; th != &thinkercap; )
    {
        thinker_t *next = th->next;
        if (th->function.acp1 == (actionf_p1) P_MobjThinker)
            P_RemoveMobj ((mobj_t *) th);
        else
            Z_Free (th);
        th = next;
    }

    P_InitThinkers ();

    // killough 2/14/98: count number of thinkers by skipping through them
    {
        byte *sp = SaveStream.data;     // save pointer and skip header
        for (size = 1; *SaveStream.data++ == tc_mobj; size++)  // killough 2/14/98
        {                     // skip all entries, adding up count
            SaveStream.data += sizeof(mobj_t);
        }

        if (*--SaveStream.data != tc_end)
            I_Error ("Unknown tclass %i in savegame", *SaveStream.data);

        // first table entry special: 0 maps to NULL
        *(mobj_p = M_Malloc(size * sizeof *mobj_p)) = 0;   // table of pointers
        SaveStream.data = sp;           // restore save pointer
    }

    // read in saved thinkers
    for (size = 1; *SaveStream.data == tc_mobj; size++)    // killough 2/14/98
    {
        mobj_t *mobj = Z_Malloc(sizeof(mobj_t), PU_LEVEL, NULL);

        // killough 2/14/98 -- insert pointers to thinkers into table, in order:
        mobj_p[size] = mobj;

        memcpy (mobj, SaveStream.data, sizeof(mobj_t));
        SaveStream.data += sizeof(mobj_t);
        mobj->state = states + (dintptr) mobj->state;

        if (mobj->player)
            (mobj->player = &players[(dintptr) mobj->player - 1]) -> mo = mobj;

        P_SetThingPosition (mobj);
        mobj->info = &mobjinfo[mobj->type];

        // killough 2/28/98:
        // Fix for falling down into a wall after savegame loaded:
        //      mobj->floorz = mobj->subsector->sector->floorheight;
        //      mobj->ceilingz = mobj->subsector->sector->ceilingheight;
        mobj->thinker.function.acp1 = (actionf_p1) P_MobjThinker;
        P_AddThinker (&mobj->thinker);
    }

    // killough 2/14/98: adjust target and tracer fields, plus
    // lastenemy field, to correctly point to mobj thinkers.
    // NULL entries automatically handled by first table entry.
    for (th = thinkercap.next ; th != &thinkercap ; th=th->next)
    {
        ((mobj_t *) th)->target =
        mobj_p[(size_t)((mobj_t *)th)->target];

        ((mobj_t *) th)->tracer =
        mobj_p[(size_t)((mobj_t *)th)->tracer];
    }

    // phares 9/13/98: Restore sec->soundtarget pointers from indices.
    // NULL entries automatically handled by first table entry.
    for (i = 0, sec = sectors ; i < numsectors ; i++, sec++)
        sec->soundtarget = mobj_p[(size_t) sec->soundtarget];

    M_Free(mobj_p);    // free translation table
}

//
// P_ArchiveSpecials
//
enum
{
    tc_ceiling,
    tc_door,
    tc_floor,
    tc_plat,
    tc_flash,
    tc_strobe,
    tc_flicker,     //jff 8/8/98 add missing fire flicker entry
    tc_glow,
    tc_elevator,    //jff 2/22/98 new elevator type thinker                
    tc_scroll,      // killough 3/7/98: new scroll effect thinker
    tc_friction,    // phares 3/18/98:  new friction effect thinker
    tc_pusher,      // phares 3/22/98:  new push/pull effect thinker

    tc_endspecials
} specials_e;

//
// Things to handle:
//
// T_MoveCeiling, (ceiling_t: sector_t * swizzle), - active list
// T_VerticalDoor, (vldoor_t: sector_t * swizzle),
// T_MoveFloor, (floormove_t: sector_t * swizzle),
// T_LightFlash, (lightflash_t: sector_t * swizzle),
// T_StrobeFlash, (strobe_t: sector_t *),
// T_Glow, (glow_t: sector_t *),
// T_PlatRaise, (plat_t: sector_t *), - active list
//
void P_ArchiveSpecials(void)
{
    thinker_t *th;
    ceiling_t *ceiling;
    vldoor_t *door;
    floormove_t *floor;
    plat_t *plat;
    lightflash_t *flash;
    strobe_t *strobe;
    glow_t *glow;
    int i;

    // save off the current thinkers
    for (th = thinkercap.next; th != &thinkercap; th = th->next)
    {
        if (th->function.acv == (actionf_v)NULL)
        {
            for (i = 0; i < MAXCEILINGS; i++)
                if (activeceilings[i] == (ceiling_t *)th)
                    break;

            if (i < MAXCEILINGS)
            {
                Stream_WriteByte(&SaveStream, tc_ceiling);
                ceiling = (ceiling_t *)SaveStream.data;
                memcpy(ceiling, th, sizeof(*ceiling));
                SaveStream.data += sizeof(*ceiling);
                ceiling->sector = (sector_t *)(ceiling->sector - sectors);
            }
            continue;
        }

        if (th->function.acp1 == (actionf_p1)T_MoveCeiling)
        {
            Stream_WriteByte(&SaveStream, tc_ceiling);
            ceiling = (ceiling_t *)SaveStream.data;
            memcpy(ceiling, th, sizeof(*ceiling));
            SaveStream.data += sizeof(*ceiling);
            ceiling->sector = (sector_t *)(ceiling->sector - sectors);
            continue;
        }

        if (th->function.acp1 == (actionf_p1)T_VerticalDoor)
        {
            Stream_WriteByte(&SaveStream, tc_door);
            door = (vldoor_t *)SaveStream.data;
            memcpy(door, th, sizeof(*door));
            SaveStream.data += sizeof(*door);
            door->sector = (sector_t *)(door->sector - sectors);
            continue;
        }

        if (th->function.acp1 == (actionf_p1)T_MoveFloor)
        {
            Stream_WriteByte(&SaveStream, tc_floor);
            floor = (floormove_t *)SaveStream.data;
            memcpy(floor, th, sizeof(*floor));
            SaveStream.data += sizeof(*floor);
            floor->sector = (sector_t *)(floor->sector - sectors);
            continue;
        }

        if (th->function.acp1 == (actionf_p1)T_PlatRaise)
        {
            Stream_WriteByte(&SaveStream, tc_plat);
            plat = (plat_t *)SaveStream.data;
            memcpy(plat, th, sizeof(*plat));
            SaveStream.data += sizeof(*plat);
            plat->sector = (sector_t *)(plat->sector - sectors);
            continue;
        }

        if (th->function.acp1 == (actionf_p1)T_LightFlash)
        {
            Stream_WriteByte(&SaveStream, tc_flash);
            flash = (lightflash_t *)SaveStream.data;
            memcpy(flash, th, sizeof(*flash));
            SaveStream.data += sizeof(*flash);
            flash->sector = (sector_t *)(flash->sector - sectors);
            continue;
        }

        if (th->function.acp1 == (actionf_p1)T_StrobeFlash)
        {
            Stream_WriteByte(&SaveStream, tc_strobe);
            strobe = (strobe_t *)SaveStream.data;
            memcpy(strobe, th, sizeof(*strobe));
            SaveStream.data += sizeof(*strobe);
            strobe->sector = (sector_t *)(strobe->sector - sectors);
            continue;
        }

        if (th->function.acp1 == (actionf_p1)T_Glow)
        {
            Stream_WriteByte(&SaveStream, tc_glow);
            glow = (glow_t *)SaveStream.data;
            memcpy(glow, th, sizeof(*glow));
            SaveStream.data += sizeof(*glow);
            glow->sector = (sector_t *)(glow->sector - sectors);
            continue;
        }

        // killough 3/7/98: Scroll effect thinkers
        if (th->function.acp1 == (actionf_p1) T_Scroll)
        {
            Stream_WriteByte(&SaveStream, tc_scroll);
            memcpy (SaveStream.data, th, sizeof(scroll_t));
            SaveStream.data += sizeof(scroll_t);
            continue;
        }

        // phares 3/18/98: Friction effect thinkers
        if (th->function.acp1 == (actionf_p1) T_Friction)
        {
            Stream_WriteByte(&SaveStream, tc_friction);
            memcpy (SaveStream.data, th, sizeof(friction_t));
            SaveStream.data += sizeof(friction_t);
            continue;
        }

        // phares 3/22/98: Push/Pull effect thinkers
        if (th->function.acp1 == (actionf_p1) T_Pusher)
        {
            Stream_WriteByte(&SaveStream, tc_pusher);
            memcpy (SaveStream.data, th, sizeof(pusher_t));
            SaveStream.data += sizeof(pusher_t);
            continue;
        }
    }

    // add a terminating marker
    Stream_WriteByte(&SaveStream, tc_endspecials);
}

//
// P_UnArchiveSpecials
//
void P_UnArchiveSpecials(void)
{
    byte tclass;
    ceiling_t *ceiling;
    vldoor_t *door;
    floormove_t *floor;
    plat_t *plat;
    lightflash_t *flash;
    strobe_t *strobe;
    glow_t *glow;

    // read in saved thinkers
    while (1)
    {
        tclass = Stream_ReadByte(&SaveStream);
        switch (tclass)
        {
        case tc_endspecials:
            return; // end of list

        case tc_ceiling:
            ceiling = (ceiling_t *)Z_Malloc(sizeof(*ceiling), PU_LEVEL, NULL);
            memcpy(ceiling, SaveStream.data, sizeof(*ceiling));
            SaveStream.data += sizeof(*ceiling);
            ceiling->sector = &sectors[(dintptr)ceiling->sector];
            ceiling->sector->specialdata = ceiling;

            if (ceiling->thinker.function.acp1)
                ceiling->thinker.function.acp1 = (actionf_p1)T_MoveCeiling;

            P_AddThinker(&ceiling->thinker);
            P_AddActiveCeiling(ceiling);
            break;

        case tc_door:
            door = (vldoor_t *)Z_Malloc(sizeof(*door), PU_LEVEL, NULL);
            memcpy(door, SaveStream.data, sizeof(*door));
            SaveStream.data += sizeof(*door);
            door->sector = &sectors[(dintptr)door->sector];
            door->sector->specialdata = door;
            door->thinker.function.acp1 = (actionf_p1)T_VerticalDoor;
            P_AddThinker(&door->thinker);
            break;

        case tc_floor:
            floor = (floormove_t *)Z_Malloc(sizeof(*floor), PU_LEVEL, NULL);
            memcpy(floor, SaveStream.data, sizeof(*floor));
            SaveStream.data += sizeof(*floor);
            floor->sector = &sectors[(dintptr)floor->sector];
            floor->sector->specialdata = floor;
            floor->thinker.function.acp1 = (actionf_p1)T_MoveFloor;
            P_AddThinker(&floor->thinker);
            break;

        case tc_plat:
            plat = (plat_t *)Z_Malloc(sizeof(*plat), PU_LEVEL, NULL);
            memcpy(plat, SaveStream.data, sizeof(*plat));
            SaveStream.data += sizeof(*plat);
            plat->sector = &sectors[(dintptr)plat->sector];
            plat->sector->specialdata = plat;

            if (plat->thinker.function.acp1)
                plat->thinker.function.acp1 = (actionf_p1)T_PlatRaise;

            P_AddThinker(&plat->thinker);
            P_AddActivePlat(plat);
            break;

        case tc_flash:
            flash = (lightflash_t *)Z_Malloc(sizeof(*flash), PU_LEVEL, NULL);
            memcpy(flash, SaveStream.data, sizeof(*flash));
            SaveStream.data += sizeof(*flash);
            flash->sector = &sectors[(dintptr)flash->sector];
            flash->thinker.function.acp1 = (actionf_p1)T_LightFlash;
            P_AddThinker(&flash->thinker);
            break;

        case tc_strobe:
            strobe = (strobe_t *)Z_Malloc(sizeof(*strobe), PU_LEVEL, NULL);
            memcpy(strobe, SaveStream.data, sizeof(*strobe));
            SaveStream.data += sizeof(*strobe);
            strobe->sector = &sectors[(dintptr)strobe->sector];
            strobe->thinker.function.acp1 = (actionf_p1)T_StrobeFlash;
            P_AddThinker(&strobe->thinker);
            break;

        case tc_glow:
            glow = (glow_t *)Z_Malloc(sizeof(*glow), PU_LEVEL, NULL);
            memcpy(glow, SaveStream.data, sizeof(*glow));
            SaveStream.data += sizeof(*glow);
            glow->sector = &sectors[(dintptr)glow->sector];
            glow->thinker.function.acp1 = (actionf_p1)T_Glow;
            P_AddThinker(&glow->thinker);
            break;
                 
        case tc_scroll:       // killough 3/7/98: scroll effect thinkers
        {
            scroll_t *scroll = Z_Malloc (sizeof(scroll_t), PU_LEVEL, NULL);
            memcpy (scroll, SaveStream.data, sizeof(scroll_t));
            SaveStream.data += sizeof(scroll_t);
            scroll->thinker.function.acp1 = (actionf_p1) T_Scroll;
            P_AddThinker(&scroll->thinker);
            break;
        }

        case tc_friction:   // phares 3/18/98: new friction effect thinkers
        {
            friction_t *friction = Z_Malloc (sizeof(friction_t), PU_LEVEL, NULL);
            memcpy (friction, SaveStream.data, sizeof(friction_t));
            SaveStream.data += sizeof(friction_t);
            friction->thinker.function.acp1 = (actionf_p1) T_Friction;
            P_AddThinker(&friction->thinker);
            break;
        }

        case tc_pusher:   // phares 3/22/98: new Push/Pull effect thinkers
        {
            pusher_t *pusher = Z_Malloc (sizeof(pusher_t), PU_LEVEL, NULL);
            memcpy (pusher, SaveStream.data, sizeof(pusher_t));
            SaveStream.data += sizeof(pusher_t);
            pusher->thinker.function.acp1 = (actionf_p1) T_Pusher;
            pusher->source = P_GetPushThing(pusher->affectee);
            P_AddThinker(&pusher->thinker);
            break;
        }

        default:
            I_Error("P_UnarchiveSpecials:Unknown tclass %i "
                    "in savegame",
                    tclass);
        }
    }
}
