//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
//-----------------------------------------------------------------------------

#include <sys/time.h>
#include <SDL2/SDL.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <libgen.h>
#include "i_system.h"
#include "i_video.h"
#include "doomdef.h"
#include "doomtype.h"
#include "g_game.h"
#include "m_alloc.h"
#include "d_net.h"
#include "version.h"
#include "m_misc.h"
#include "c_console.h"
#include "i_sound.h"

extern SDL_Window *window;

int mb_used = 64;

void I_Tactile(int on,
               int off,
               int total)
{
    // UNUSED.
    on = off = total = 0;
}

ticcmd_t emptycmd;
ticcmd_t *I_BaseTiccmd(void)
{
    return &emptycmd;
}

int I_GetHeapSize(void)
{
    return mb_used * 1024 * 1024;
}

byte *I_ZoneBase(int *size)
{
    *size = mb_used * 1024 * 1024;
    return (byte *)M_Malloc(*size);
}

static boolean DirectoryExists(const char *path) 
{
    DIR *dir;

    dir = opendir(path);
    if (dir)
    {
        closedir(dir);
        return true;
    }

    return false;
}

const char *I_GetConfigDir()
{
    static boolean inited = false;
    static char buffer[256];

    if (!inited)
    {
        const char *home;
        const char *xdg_config_home;
        
        memset(buffer, 0, sizeof(buffer));

        xdg_config_home = getenv("XDG_CONFIG_HOME");
        if (!xdg_config_home)
        {
            home = getenv("HOME");
            if (!home)
            {
                I_Error("Failed to get config directory");
            }

            strcat(buffer, home);
            
            buffer[strlen(buffer)] = '/';
            strcat(buffer, ".config");

            if (!DirectoryExists(buffer))
                if (mkdir(buffer, S_IRWXU) == -1)
                    I_Error("Failed to create directory %s", buffer);
        }
        else
        {
            strcat(buffer, xdg_config_home);
        }
        

        buffer[strlen(buffer)] = '/';
        strcat(buffer, PACKAGE_ID);

        if (!DirectoryExists(buffer))
            if (mkdir(buffer, S_IRWXU) == -1)
                I_Error("Failed to create config directory %s\n", buffer);

        inited = true;
    }

    return buffer;
}

const char *I_GetExeDir()
{
    static boolean inited = false;
    static char buffer[256];
    static char *output = NULL;
    
    if (!inited)
    {
        memset(buffer, 0, 256);
        
        readlink("/proc/self/exe", buffer, 256);
        output = dirname(buffer);
        
        inited = true;
    }
    
    return output;
}

//
// I_GetTime
// returns time in 1/70th second tics
//
int I_GetTime(void)
{
    struct timeval tp;
    struct timezone tzp;
    int newtics;
    static int basetime = 0;

    gettimeofday(&tp, &tzp);
    if (!basetime)
        basetime = tp.tv_sec;
    newtics = (tp.tv_sec - basetime) * TICRATE + tp.tv_usec * TICRATE / 1000000;
    return newtics;
}

double I_GetMsTime()
{
    Uint64 qpc, qpf;
    double v;

    qpc = SDL_GetPerformanceCounter();
    qpf = SDL_GetPerformanceFrequency();

    v = (((double)qpc) / ((double)qpf));
    v *= 1000.0;
    return v;
}

void I_Init()
{
    SDL_version linked;

    SDL_Init(SDL_INIT_EVERYTHING);

    SDL_GetVersion(&linked);
    Printf("Using SDL v%d.%d.%d\n", linked.major, linked.minor, linked.patch);
    I_InitSound();
}

void I_Quit()
{
    D_QuitNetGame();
    I_ShutdownSound();
    M_SaveDefaults();
    I_ShutdownGraphics();

    SDL_Quit();

    exit(0);
}

void I_WaitVBL(int count)
{
    usleep((1000 / 60) * count); /* [Petteri] Is this a good idea? */
}

byte *I_AllocLow(int length)
{
    byte *mem;

    mem = (byte *)M_Malloc(length);
    memset(mem, 0, length);
    return mem;
}

//
// I_Error
//
extern boolean demorecording;

void I_Error(char *error, ...)
{
    char buf[1024];
    va_list argptr;

    // Message first.
    va_start(argptr, error);
    vsnprintf(buf, sizeof(buf), error, argptr);
    va_end(argptr);

    fprintf(stderr, "Error: %s\n", buf);
    fflush(stderr);

    SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_ERROR, "Fatal Error!", buf, window);

    // Shutdown. Here might be other errors.
    if (demorecording)
        G_CheckDemoStatus();

    D_QuitNetGame();
    I_ShutdownGraphics();

    /* [Petteri] We'd better shut down the sound system as well: */
    I_ShutdownSound();

    SDL_Quit();

    exit(-1);
}

#define ALTAIR_KEYMAP(k, s)                                            \
    case k: return s;

const char *I_MapKeyToName(event_t *event)
{
    static char name[64];

    memset(name, 0, sizeof(name));

    switch (event->type)
    {
        case ev_keydown:
        case ev_keyup:
        {
            size_t i, len;
            int key = event->data2;
            
			switch (key)
            {
				ALTAIR_KEYMAP(SDLK_BACKSPACE, "backspace");
				ALTAIR_KEYMAP(SDLK_TAB, "tab");
				ALTAIR_KEYMAP(SDLK_CLEAR, "clear");
				ALTAIR_KEYMAP(SDLK_RETURN, "enter");
				ALTAIR_KEYMAP(SDLK_LSHIFT, "shift");
				ALTAIR_KEYMAP(SDLK_RSHIFT, "shift");
				ALTAIR_KEYMAP(SDLK_LCTRL, "ctrl");
				ALTAIR_KEYMAP(SDLK_RCTRL, "ctrl");
				ALTAIR_KEYMAP(SDLK_MENU, "alt");
				ALTAIR_KEYMAP(SDLK_LALT, "alt");
				ALTAIR_KEYMAP(SDLK_RALT, "alt");
				ALTAIR_KEYMAP(SDLK_PAUSE, "pause");
				ALTAIR_KEYMAP(SDLK_ESCAPE, "escape");
				ALTAIR_KEYMAP(SDLK_SPACE, "space");
				ALTAIR_KEYMAP(SDLK_PRIOR, "pgup");
				ALTAIR_KEYMAP(SDLK_END, "end");
				ALTAIR_KEYMAP(SDLK_HOME, "home");
				ALTAIR_KEYMAP(SDLK_LEFT, "left");
				ALTAIR_KEYMAP(SDLK_UP, "up");
				ALTAIR_KEYMAP(SDLK_RIGHT, "right");
				ALTAIR_KEYMAP(SDLK_DOWN, "down");
				ALTAIR_KEYMAP(SDLK_SELECT, "select");
				ALTAIR_KEYMAP(SDLK_INSERT, "insert");
				ALTAIR_KEYMAP(SDLK_DELETE, "delete");
				ALTAIR_KEYMAP(SDLK_F1, "f1");
				ALTAIR_KEYMAP(SDLK_F2, "f2");
				ALTAIR_KEYMAP(SDLK_F3, "f3");
				ALTAIR_KEYMAP(SDLK_F4, "f4");
				ALTAIR_KEYMAP(SDLK_F5, "f5");
				ALTAIR_KEYMAP(SDLK_F6, "f6");
				ALTAIR_KEYMAP(SDLK_F7, "f7");
				ALTAIR_KEYMAP(SDLK_F8, "f8");
				ALTAIR_KEYMAP(SDLK_F9, "f9");
				ALTAIR_KEYMAP(SDLK_F10, "f10");
				ALTAIR_KEYMAP(SDLK_F11, "f11");
				ALTAIR_KEYMAP(SDLK_F12, "f12");
				ALTAIR_KEYMAP(SDLK_F13, "f13");
				ALTAIR_KEYMAP(SDLK_F14, "f14");
				ALTAIR_KEYMAP(SDLK_F15, "f15");
				ALTAIR_KEYMAP(SDLK_F16, "f16");
				ALTAIR_KEYMAP(SDLK_F17, "f17");
				ALTAIR_KEYMAP(SDLK_F18, "f18");
				ALTAIR_KEYMAP(SDLK_F19, "f19");
				ALTAIR_KEYMAP(SDLK_F20, "f20");
				ALTAIR_KEYMAP(SDLK_F21, "f21");
				ALTAIR_KEYMAP(SDLK_F22, "f22");
				ALTAIR_KEYMAP(SDLK_F23, "f23");
				ALTAIR_KEYMAP(SDLK_F24, "f24");
            }

            strncpy(name, SDL_GetKeyName(event->data2), sizeof(name));

            len = strlen(name);

            for (i = 0; i < len; i++)
            {
                name[i] = (char)tolower(name[i]);
            }

            return name;
        }
    }

    return NULL;
}
