//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
//-----------------------------------------------------------------------------

#include <GL/glew.h>
#include <SDL2/SDL.h>
#include <time.h>
#include "i_video.h"
#include "i_system.h"
#include "v_video.h"
#include "multires.h"
#include "version.h"
#include "hginfo.h"
#include "c_cvar.h"
#include "c_console.h"
#include "d_event.h"
#include "d_main.h"
#include "i_sound.h"
#include "w_wad.h"
#include "z_zone.h"
#include "m_alloc.h"

SDL_Window *window = NULL;

cvar_t *vid_showfps,
    *vid_vsync;

cvar_t *vid_width;
cvar_t *vid_height;
cvar_t *vid_fullscreen;

extern cvar_t *vid_fullscreen;

static int fps = 0;
static int fpsFrames = 0;
static double lastFpsTime = 0.0;
static double DeltaTime = 0.0;

static GLuint ShaderProgram;
static GLuint VertexBuffer, VertexArrays;
static GLuint ScreenTexture, PaletteTexture;

extern byte gammatable[256];

static void CheckGlShaderCompile(GLuint shader);
static void CalculateOrthoMatrix(float output[4][4], float bottom, float top, float left, float right, float near, float far);

int TranslateKey(SDL_Keycode k)
{
    if (k >= ' ' && k <= 'z')
		return k;

#define K(a, b) \
    case a:     \
        return b;
    switch (k)
    {
        K(SDLK_LEFT, KEY_LEFTARROW);
        K(SDLK_RIGHT, KEY_RIGHTARROW);
        K(SDLK_UP, KEY_UPARROW);
        K(SDLK_DOWN, KEY_DOWNARROW);
        K(SDLK_BACKSPACE, KEY_BACKSPACE);
        K(SDLK_TAB, KEY_TAB);
        K(SDLK_RETURN, KEY_ENTER);
        K(SDLK_RSHIFT, KEY_RSHIFT);
        K(SDLK_RCTRL, KEY_RCTRL);
        K(SDLK_LSHIFT, KEY_RSHIFT);
        K(SDLK_LCTRL, KEY_RCTRL);
        K(SDLK_MENU, KEY_RALT);
        K(SDLK_PAUSE, KEY_PAUSE);
        K(SDLK_ESCAPE, KEY_ESCAPE);
        K(SDLK_SPACE, ' ');
        K(SDLK_DELETE, KEY_BACKSPACE);
        K(SDLK_PLUS, '+');
        K(SDLK_MINUS, KEY_MINUS);
        K(SDLK_COMMA, ',');
        K(SDLK_PERIOD, '.');
        K(SDLK_F1, KEY_F1);
        K(SDLK_F2, KEY_F2);
        K(SDLK_F3, KEY_F3);
        K(SDLK_F4, KEY_F4);
        K(SDLK_F5, KEY_F5);
        K(SDLK_F6, KEY_F6);
        K(SDLK_F7, KEY_F7);
        K(SDLK_F8, KEY_F8);
        K(SDLK_F9, KEY_F9);
        K(SDLK_F10, KEY_F10);
        K(SDLK_F11, KEY_F11);
        K(SDLK_F12, KEY_F12);
        K(SDLK_BACKQUOTE, KEY_GRAVE);
        K(SDLK_SEMICOLON, KEY_SEMICOLON);
        K(SDLK_PAGEUP, KEY_PGUP);
        K(SDLK_PAGEDOWN, KEY_PGDN);
    }

    return 0;
}

void I_GetEvent()
{
    event_t event;
    SDL_Event ev;
    while (SDL_PollEvent(&ev)) 
    {
        switch (ev.type)
        {
            case SDL_QUIT:
            {
                SDL_HideWindow(window);
                I_Quit();

                break;
            }

            case SDL_KEYDOWN:
            case SDL_KEYUP:
            {
                event.type = (ev.type == SDL_KEYDOWN ? ev_keydown : ev_keyup);
                event.data1 = TranslateKey(ev.key.keysym.sym);
                event.data2 = ev.key.keysym.sym;
                event.data3 = ev.key.keysym.scancode;
                
                D_PostEvent(&event);

                break;
            }

            case SDL_TEXTINPUT:
            {
                size_t len, i;

                len = strlen(ev.text.text);

                for (i = 0; i < len; i++)
                {
                    char c = ev.text.text[i];

                    event.type = ev_type;
                    event.data1 = (int)c;
                    event.data2 = ev.key.keysym.sym;
                    event.data3 = ev.key.keysym.scancode;
                    
                    D_PostEvent(&event);
                }

                break;
            }
        }
    }
}

void I_StartTyping()
{
    SDL_StartTextInput();
}

void I_StopTyping()
{
    SDL_StopTextInput();
}

//
// I_StartTic
//
void I_StartTic(void)
{
    I_GetEvent();
}

/* Stuff for multiresolution support: */
int SCREENWIDTH;
int SCREENHEIGHT;
int weirdaspect;
static boolean IsGlInited = false;

//
// I_StartFrame
//
void I_StartFrame(void)
{
    // er?
}

//
// I_UpdateNoBlit
//
void I_UpdateNoBlit(void)
{
    // what is this?
}

void I_InitGraphics(void)
{
    char title[64];

    sprintf(title, "%s v%s (%s) r%s", PACKAGE_NAME, PACKAGE_VERSION, __DATE__, HG_REVISION_STR);

    window = SDL_CreateWindow(title, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, SCREENWIDTH, SCREENHEIGHT, SDL_WINDOW_HIDDEN|SDL_WINDOW_OPENGL);
    if (!window)
    {
        I_Error("Failed to create SDL window: %s.", SDL_GetError());
    }

    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 3);
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 3);
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_CORE);
    SDL_GLContext ctx = SDL_GL_CreateContext(window);
    if (!ctx)
    {
        I_Error("Failed to create OpenGL context: %s.", SDL_GetError());
    }
    SDL_GL_MakeCurrent(window, ctx);

    glewExperimental = GL_TRUE;
    GLenum err = glewInit();
    if (err != GLEW_OK)
    {
        I_Error("Failed to initialise GLEW: %d", err);
    }

    IsGlInited = true;

    Printf("I_InitGraphics:   Vendor: %s\n", glGetString(GL_VENDOR));
    Printf("I_InitGraphics: Renderer: %s\n", glGetString(GL_RENDERER));
    Printf("I_InitGraphics:  Version: %s\n", glGetString(GL_VERSION));

    // Now create the GL objects
    {
        ShaderProgram = glCreateProgram();

        GLuint frag = glCreateShader(GL_FRAGMENT_SHADER);
        GLuint vert = glCreateShader(GL_VERTEX_SHADER);

        char *code;
        int num, len;
        int successful;

        // Load fragment shader
        num = W_GetNumForName("ALGLFS");
        len = W_LumpLength(num);
        code = W_CacheLumpNum(num, PU_CACHE);
        glShaderSource(frag, 1, &code, &len);
        glCompileShader(frag);
        CheckGlShaderCompile(frag);

        // Load vertex shader
        num = W_GetNumForName("ALGLVS");
        len = W_LumpLength(num);
        code = W_CacheLumpNum(num, PU_CACHE);
        glShaderSource(vert, 1, &code, &len);
        glCompileShader(vert);
        CheckGlShaderCompile(vert);

        // Attach them to the program
        glAttachShader(ShaderProgram, frag);
        glAttachShader(ShaderProgram, vert);

        // Link it
        glLinkProgram(ShaderProgram);
        glGetProgramiv(ShaderProgram, GL_LINK_STATUS, &successful);
        if (!successful)
        {
            char buf[1024];
            glGetProgramInfoLog(ShaderProgram, sizeof(buf), NULL, buf);
            I_Error("Failed to link shader program %d:\n%s", ShaderProgram, buf);
        }

        // Detach and remove the shaders
        glDetachShader(ShaderProgram, vert);
        glDeleteShader(vert);
        glDetachShader(ShaderProgram, frag);
        glDeleteShader(frag);

        // and bind our program
        glUseProgram(ShaderProgram);
    }

    // Create VBO and VAO
    {
        float sw = (float)SCREENWIDTH;
        float sh = (float)SCREENHEIGHT;

        glGenBuffers(1, &VertexBuffer);
        glGenVertexArrays(1, &VertexArrays);

        glBindVertexArray(VertexArrays);
        glBindBuffer(GL_ARRAY_BUFFER, VertexBuffer);

        // position
        glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 4 * sizeof(float), (void *)(0 * sizeof(float)));
        glEnableVertexAttribArray(0);
        // texCoord
        glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 4 * sizeof(float), (void *)(2 * sizeof(float)));
        glEnableVertexAttribArray(1);

        float verts[6 * 4] =
        {
                0.0f,   0.0f,   0.0f,   0.0f,
                0.0f,   sh,     0.0f,   1.0f,
                sw,     sh,     1.0f,   1.0f,
                sw,     sh,     1.0f,   1.0f,
                sw,     0.0f,   1.0f,   0.0f,
                0.0f,   0.0f,   0.0f,   0.0f,
        };

        glBufferData(GL_ARRAY_BUFFER, sizeof verts, verts, GL_STATIC_DRAW);
    }

    // Create the screen texture
    {
        glGenTextures(1, &ScreenTexture);
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, ScreenTexture);

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

        glTexImage2D(GL_TEXTURE_2D, 0, GL_RED, SCREENWIDTH, SCREENHEIGHT, 0, GL_RED, GL_UNSIGNED_BYTE, NULL);
    }

    // Create the palette texture
    {
        glGenTextures(1, &PaletteTexture);
        glActiveTexture(GL_TEXTURE1);
        glBindTexture(GL_TEXTURE_2D, PaletteTexture);

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 256, 1, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
    }

    // Bind the textures to their appropriate shader samplers
    {
        GLuint screenLoc = glGetUniformLocation(ShaderProgram, "u_screen");
        GLuint paletteLoc = glGetUniformLocation(ShaderProgram, "u_palette");

        glUniform1i(screenLoc, 0);
        glUniform1i(paletteLoc, 1);
    }

    SDL_ShowWindow(window);
}

void I_ShutdownGraphics(void)
{
    Printf("Shutting down SDL.\n");

    if (IsGlInited)
    {
        glDeleteTextures(1, &ScreenTexture);
        glDeleteTextures(1, &PaletteTexture);
        glDeleteBuffers(1, &VertexBuffer);
        glDeleteVertexArrays(1, &VertexArrays);
        glDeleteProgram(ShaderProgram);
    }

    SDL_DestroyWindow(window);
}

static void CheckGlShaderCompile(GLuint shader)
{
    int status;
    glGetShaderiv(shader, GL_COMPILE_STATUS, &status);
    if (!status)
    {
        char buf[1024];
        glGetShaderInfoLog(shader, sizeof(buf), NULL, buf);
        I_Error("Failed to compile shader %d:\n%s", shader, buf);
    }
}

void I_AllocScreen(screen_t *screen, int width, int height, int depth)
{
    memset(screen, 0, sizeof(screen_t));

    screen->width = width;
    screen->height = height;
    screen->depth = depth;
    screen->platdata = M_Malloc(width * height * depth);
    memset(screen->platdata, 0, width * depth * height);
}

void I_FreeScreen(screen_t *screen)
{
    M_Free(screen->platdata);
    memset(screen, 0, sizeof(screen_t));
}

boolean I_LockScreen(screen_t *screen)
{
    if (screen->locked)
    {
        screen->locked++;
        return 1;
    }

    if (screen == NULL || screen->platdata == NULL)
    {
        return 0;
    }

    screen->locked++;
    screen->pitch = screen->width * screen->depth;
    screen->data = screen->platdata;

    return 1;
}

boolean I_UnlockScreen(screen_t *screen)
{
    if (!screen->locked)
    {
        return 1;
    }

    if (--screen->locked)
    {
        return 1;
    }

    screen->pitch = 0;
    screen->locked = 0;
    screen->data = NULL;

    return 1;
}

void I_FinishUpdate(void)
{
    double start, end;
    float matrix[4][4];
    
    I_DrawSoundOverlay();

    if (vid_showfps->value)
    {
        char str[128];
        int x, y;

        sprintf(str, "%d FPS (%f ms)", fps, DeltaTime);

        x = SCREENWIDTH - Con_StringWidth(str);
        y = 0;

        Con_DrawString(x, y, str);
    }

    start = I_GetMsTime();

    glViewport(0, 0, SCREENWIDTH, SCREENHEIGHT);

    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);

    glActiveTexture(GL_TEXTURE0);
    I_LockScreen(&screens[0]);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RED, SCREENWIDTH, SCREENHEIGHT, 0, GL_RED, GL_UNSIGNED_BYTE, screens[0].data);
    I_UnlockScreen(&screens[0]);

    CalculateOrthoMatrix(matrix, (float)SCREENHEIGHT, 0.0f, 0.0f, (float)SCREENWIDTH, -1.0f, 1.0f);
    GLuint matrixLoc = glGetUniformLocation(ShaderProgram, "u_projection");
    glUniformMatrix4fv(matrixLoc, 1, GL_FALSE, matrix);

    glDrawArrays(GL_TRIANGLES, 0, 6);

    SDL_GL_SwapWindow(window);

    end = I_GetMsTime();
    DeltaTime = end - start;
    fpsFrames++;
    if (end - lastFpsTime >= 1000.0)
    {
        fps = fpsFrames;
        fpsFrames = 0;
        lastFpsTime = end;
    }
}

void CalculateOrthoMatrix(float output[4][4], float bottom, float top, float left, float right, float near, float far)
{
    output[0][0] = 2 / (right - left);
    output[0][1] = 0;
    output[0][2] = 0;
    output[0][3] = 0;

    output[1][0] = 0;
    output[1][1] = 2 / (top - bottom);
    output[1][2] = 0;
    output[1][3] = 0;

    output[2][0] = 0;
    output[2][1] = 0;
    output[2][2] = -2 / (far - near);
    output[2][3] = 0;

    output[3][0] = -(right + left) / (right - left);
    output[3][1] = -(top + bottom) / (top - bottom);
    output[3][2] = -(far + near) / (far - near);
    output[3][3] = 1;
}

void I_ReadScreen(byte *scr)
{
    I_LockScreen(&screens[0]);
    memcpy(scr, screens[0].data, screens[0].height * screens[0].pitch);
    I_UnlockScreen(&screens[0]);
}

void I_SetPalette(byte *pal)
{
    int i;
    byte tex[256 * 3];

    if (!PaletteTexture)
    {
        return;
    }

    for (int i = 0; i < 256; i++)
    {
        byte *src = pal + (i * 3);
        byte *dst = tex + (i * 3);

        dst[0] = gammatable[src[0]];
        dst[1] = gammatable[src[1]];
        dst[2] = gammatable[src[2]];
    }

    glActiveTexture(GL_TEXTURE1);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 256, 1, 0, GL_RGB, GL_UNSIGNED_BYTE, tex);
}

void I_Screenshot()
{
    char filename[512];
    char timestamp[64];
    static char lastTimestamp[64];
    static unsigned counter = 0;
    SDL_Surface *sshot, *flipped;

    counter++;

    // Generate filename
    {
        time_t rawtime;
        struct tm *info;
        time(&rawtime);
        info = localtime(&rawtime);
        
        memset(timestamp, 0, 64);
        strftime(timestamp, 64, "%Y-%m-%d_%H-%M-%S", info);

        if (strcmp(timestamp, lastTimestamp))
        {
            memcpy(lastTimestamp, timestamp, 64);
            counter = 0;
        }

        sprintf(filename, "%s/Screenshot_%s_%s_%u.bmp", I_GetConfigDir(), PACKAGE_ID, timestamp, counter);
    }

    uint32_t rmask = 0x000000FF;
    uint32_t gmask = 0x0000FF00;
    uint32_t bmask = 0x00FF0000;
    uint32_t amask = 0xFF000000;

    sshot = SDL_CreateRGBSurface(0, SCREENWIDTH, SCREENHEIGHT, 32, rmask, gmask, bmask, amask);
    SDL_LockSurface(sshot);

    glReadPixels(0, 0, SCREENWIDTH, SCREENHEIGHT, GL_RGBA, GL_UNSIGNED_BYTE, sshot->pixels);

    // We need to flip our buffer because OpenGL
    flipped = SDL_CreateRGBSurface(0, SCREENWIDTH, SCREENHEIGHT, 32, rmask, gmask, bmask, amask);
    SDL_LockSurface(flipped);

    int pxlength = sshot->pitch * (sshot->h - 1);
    byte *pixels = sshot->pixels + pxlength;
    byte *fpixels = flipped->pixels;

    for (int line = 0; line < sshot->h; ++line) {
        memcpy(fpixels, pixels, sshot->pitch);

        pixels -= sshot->pitch;
        fpixels += sshot->pitch;
    }

    SDL_UnlockSurface(flipped);
    SDL_UnlockSurface(sshot);

    SDL_SaveBMP(flipped, filename);
    SDL_FreeSurface(flipped);
    SDL_FreeSurface(sshot);

    Printf("Saved screenshot as %s\n", filename);
}

void I_UpdateWindowTitle(const char *fmt, ...)
{
    va_list args;
    char buf[256];
    char title[512];
    
    va_start (args, fmt);
    vsprintf (buf, fmt, args);
    va_end (args);

    if (buf[0] != '\0')
        sprintf(title, "%s -- %s v%s (%s) r%s", buf, PACKAGE_NAME, PACKAGE_VERSION, __DATE__, HG_REVISION_STR);
    else
        sprintf(title, "%s v%s (%s) r%s", PACKAGE_NAME, PACKAGE_VERSION, __DATE__, HG_REVISION_STR);

    SDL_SetWindowTitle(window, title);
}
