//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// Copyright 1993-1997 id Software
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
//-----------------------------------------------------------------------------

#include "v_video.h"
#include "m_swap.h"
#include "i_video.h"

byte *V_TranslationTables;
byte *V_TranslationTable;

void V_DrawColumn(byte *source, byte *dest, int count, int pitch);
void V_DrawColumnTranslated(byte *source, byte *dest, int count, int pitch);

void V_DrawColumnStretched(byte *source, byte *dest, int count, int pitch, int yinc);
void V_DrawColumnTranslatedStretched(byte *source, byte *dest, int count, int pitch, int yinc);

ColumnDrawer ColumnDrawers[D_NumColumnDrawers] =
{
    V_DrawColumn,
    V_DrawColumnTranslated
};

ColumnDrawerStretched StretchedColumnDrawers[D_NumColumnDrawers] =
{
    V_DrawColumnStretched,
    V_DrawColumnTranslatedStretched
};

///////////////////////////////////////////////////////////////////////////////
// Column drawing functions
///////////////////////////////////////////////////////////////////////////////
void V_DrawColumn(byte *source, byte *dest, int count, int pitch)
{
    while (count--)
    {
        *dest = *source++;
        dest += pitch;
    }
}

void V_DrawColumnTranslated(byte *source, byte *dest, int count, int pitch)
{
    while (count--)
    {
        *dest = V_TranslationTable[*source++];
        dest += pitch;
    }
}

void V_DrawColumnStretched(byte *source, byte *dest, int count, int pitch, int yinc)
{
    int c = 0;

    do {
        *dest = source[c >> 16]; 
        dest += pitch;
        c += yinc;
    } while (--count);
}

void V_DrawColumnTranslatedStretched(byte *source, byte *dest, int count, int pitch, int yinc)
{
    int c = 0;

    do {
        *dest = V_TranslationTable[source[c >> 16]]; 
        dest += pitch;
        c += yinc;
    } while (--count);
}

///////////////////////////////////////////////////////////////////////////////
// Patch drawing functions
///////////////////////////////////////////////////////////////////////////////
void V_BaseDraw(int type, int x, int y, screen_t *scrn, patch_t *patch)
{
int count;
    int col;
    column_t *column;
    byte *desttop;
    byte *dest;
    byte *source;
    int w;
    ColumnDrawer drawer;

    drawer = ColumnDrawers[type];

    y -= SHORT(patch->topoffset);
    x -= SHORT(patch->leftoffset);
#ifdef RANGECHECK
    if (x < 0 || x + SHORT(patch->width) > scrn->width || y < 0 || y + SHORT(patch->height) > scrn->height || (unsigned)scrn > 4)
    {
        fprintf(stderr, "Patch at %d,%d exceeds LFB\n", x, y);
        // No I_Error abort - what is up with TNT.WAD?
        fprintf(stderr, "V_DrawPatch8: bad patch (ignored)\n");
        return;
    }
#endif

    if (!scrn)
        V_MarkRect(x, y, SHORT(patch->width), SHORT(patch->height));

    I_LockScreen(scrn);

    col = 0;
    desttop = scrn->data + y * scrn->width + x;

    w = SHORT(patch->width);

    for (; col < w; x++, col++, desttop++)
    {
        column = (column_t *)((byte *)patch + LONG(patch->columnofs[col]));

        // step through the posts in a column
        while (column->topdelta != 0xff)
        {
            source = (byte *)column + 3;
            dest = desttop + column->topdelta * scrn->width;
            count = column->length;

            drawer(source, dest, count, scrn->pitch);

            column = (column_t *)((byte *)column + column->length + 4);
        }
    }

    I_UnlockScreen(scrn);
}

void V_BaseDrawStretched(int type, int x, int y, screen_t *scrn, patch_t *patch, int width, int height)
{
    column_t*	column; 
    byte*		desttop;
    int			colstep = 1;

    int			xinc, yinc, col, w, ymul, xmul;

    ColumnDrawerStretched drawer;

    if (width == SHORT(patch->width) && height == SHORT(patch->height))
    {
        V_DrawPatch(x, y, scrn, patch);
        return;
    }

    xinc = (SHORT(patch->width) << 16) / width;
    yinc = (SHORT(patch->height) << 16) / height;
    xmul = (width << 16) / SHORT(patch->width);
    ymul = (height << 16) / SHORT(patch->height);

    y -= (SHORT(patch->topoffset) * ymul) >> 16;
    x -= (SHORT(patch->leftoffset) * xmul) >> 16;

    I_LockScreen(scrn);

    col = 0;
    desttop = scrn->data + y * scrn->pitch + x * colstep;

    w = width * xinc;

    drawer = StretchedColumnDrawers[type];

    for (; col < w ; col += xinc, desttop += colstep)
    {
        int c = 0;

        column = (column_t *)((byte *)patch + LONG(patch->columnofs[col >> 16]));

        // step through the posts in a column
        while (column->topdelta != 0xff )
        {
            byte *source = (byte *)column + 3;
            byte *dest = desttop + (((column->topdelta * ymul)) >> 16) * scrn->pitch;
            int count = (column->length * ymul) >> 16;

            drawer(source, dest, count, scrn->pitch, yinc);

            column = (column_t *)((byte *)column + column->length + 4);
        }
    }

    I_UnlockScreen(scrn);
}

void V_BaseDrawIndirect(int type, int x, int y, screen_t *scrn, patch_t *patch)
{
    int nw, nh, dx, dy, dw, dh;

    if (scrn->width == 320 && scrn->height == 200)
    {
        V_BaseDraw(type, x, y, scrn, patch);
        return;
    }

    nw = scrn->width;
    nh = scrn->height;

    dx = (nw * x) / 320;
    dy = (nh * y) / 200;
    dw = (scrn->width * SHORT(patch->width)) / 320;
    dh = (scrn->height * SHORT(patch->height)) / 200;

    V_BaseDrawStretched(type, dx, dy, scrn, patch, dw, dh);
}