//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// Copyright 1993-1997 id Software
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
//-----------------------------------------------------------------------------

#include <ctype.h>
#include <math.h>

#include "multires.h"
#include "doomstat.h"
#include "hu_stuff.h"
#include "m_swap.h"
#include "i_system.h"
#include "r_local.h"
#include "doomdef.h"
#include "doomdata.h"
#include "m_bbox.h"
#include "v_video.h"
#include "w_wad.h"
#include "z_zone.h"
#include "c_cvar.h"
#include "i_video.h"

int dirtybox[4];

extern patch_t *hu_font[HU_FONTSIZE];
byte gammatable[256];
extern double gammalevel;

void V_DrawString(int x, int y, int cr, const char *str)
{
    int w;
    size_t i;
    char c;
    int cx;
    int cy;

    cx = x;
    cy = y;

    V_TranslationTable = V_TranslationTables + (cr * 256);

    for (i = 0; i < strlen(str); i++)
    {
        c = str[i];
        if (!c)
            break;
        if (c == '\n')
        {
            cx = x;
            cy += 12;
            continue;
        }

        c = toupper(c) - HU_FONTSTART;
        if (c < 0 || c >= HU_FONTSIZE)
        {
            cx += 4;
            continue;
        }

        w = SHORT(hu_font[c]->width);
        if (cx + w > SCREENWIDTH)
            break;
        V_DrawPatchTranslated(cx, cy, &screens[0], hu_font[c]);
        cx += w;
    }
}

void V_PalleteIndexToColor(byte index, int redness, byte *r, byte *g, byte *b)
{
    static int palnum;
    static boolean palnumCached = false;
    byte *pal, *col;

    if (!palnumCached)
    {
        palnum = W_CheckNumForName("PLAYPAL");
        palnumCached = true;
    }

    pal = (byte *)W_CacheLumpNum(palnum, PU_CACHE);
    pal += redness * 256 * 3;
    col = pal + (index * 3);

    if (r) *r = col[0];
    if (g) *g = col[1];
    if (b) *b = col[2];
}

byte V_ClosestColor(byte r, byte g, byte b)
{
    int	i;
    int	dr, dg, db;
    int	bestdistortion, distortion;
    int	bestcolor;
    static int palnum;
    static boolean palnumCached = false;
    byte	*pal;

    if (!palnumCached)
    {
        palnum = W_CheckNumForName("PLAYPAL");
        palnumCached = true;
    }

    pal = (byte *)W_CacheLumpNum(palnum, PU_CACHE);

    //
    // let any color go to 0 as a last resort
    //
    bestdistortion = 256*256*4;
    bestcolor = 0;

    for (i = 0; i <= 256; i++)
    {
        dr = r - (int)pal[0];
        dg = g - (int)pal[1];
        db = b - (int)pal[2];
        pal += 3;
        distortion = dr*dr + dg*dg + db*db;

        if (distortion < bestdistortion)
        {
            if (!distortion)
                return i;		// perfect match

            bestdistortion = distortion;
            bestcolor = i;
        }
    }

    return bestcolor;
}

void V_CalculateGamma()
{
    double lev = gammalevel;
    double inv;
    int i;

    inv = 1.0f / lev;

    for (i = 0; i < 256; i++)
    {
        gammatable[i] = (byte)(255.0 * pow (i / 255.0, inv));
    }

    I_SetPalette((byte *)W_CacheLumpName("PLAYPAL", PU_CACHE));
}

//
// V_MarkRect
//
void V_MarkRect(int x, int y, int width, int height)
{
    M_AddToBox(dirtybox, x, y);
    M_AddToBox(dirtybox, x + width - 1, y + height - 1);
}

//
// V_CopyRect
//
void V_CopyRect(int srcx, int srcy, screen_t *srcscrn, int width, int height, int destx, int desty, screen_t *destscrn)
{
    byte *src;
    byte *dest;

#ifdef RANGECHECK
    if (srcx < 0 || srcx + width > srcscrn->width || srcy < 0 || srcy + height > srcscrn->height || destx < 0 || destx + width > destscrn->width || desty < 0 || desty + height > destscrn->height || (unsigned)srcscrn > 4 || (unsigned)destscrn > 4)
    {
        I_Error("Bad V_CopyRect8");
    }
#endif
    V_MarkRect(destx, desty, width, height);

    I_LockScreen(srcscrn);
    I_LockScreen(destscrn);

    src = srcscrn->data + srcscrn->width * srcy + srcx;
    dest = destscrn->data + srcscrn->width * desty + destx;

    for (; height > 0; height--)
    {
        memcpy(dest, src, width);
        src += srcscrn->width;
        dest += destscrn->width;
    }

    I_UnlockScreen(destscrn);
    I_UnlockScreen(srcscrn);
}

//
// V_DrawBlock
// Draw a linear block of pixels into the view buffer.
//
void V_DrawBlock(int x, int y, screen_t *scrn, int width, int height, byte *src)
{
    byte *dest;
    int yy;

    V_MarkRect(x, y, width, height);

    I_LockScreen(scrn);

    dest = scrn->data + y * scrn->width + x;

    for (yy = y; yy < y + height; yy++)
    {
        // Allows patches to start drawing outside of the screen
        if (yy >= 0 && y < scrn->height)
            memcpy(dest, src, width);

        src += width;
        dest += scrn->width;
    }

    I_UnlockScreen(scrn);
}

//
// V_GetBlock
// Gets a linear block of pixels from the view buffer.
//
void V_GetBlock(int x, int y, screen_t *scrn, int width, int height, byte *dest)
{
    byte *src;

#ifdef RANGECHECK
    if (x < 0 || x + width > screen->width || y < 0 || y + height > screen->height || (unsigned)scrn > 4)
    {
        I_Error("Bad V_DrawBlock8");
    }
#endif

    I_LockScreen(scrn);
    src = scrn->data + y * scrn->width + x;

    while (height--)
    {
        memcpy(dest, src, width);
        src += scrn->width;
        dest += width;
    }

    I_UnlockScreen(scrn);
}

//
// V_Init
//
void V_Init(void)
{
    int i;

    for (i = 0; i < 2; i++)
    {
        I_AllocScreen(&screens[i], SCREENWIDTH, SCREENHEIGHT, 1);
    }
}

void V_ClearScreen(screen_t *scrn)
{
    char name[20];
    byte *src;
    byte *dest;
    int x, y;

    if (gamemode == commercial)
        strcpy(name, "GRNROCK");
    else
        strcpy(name, "FLOOR7_2");

    I_LockScreen(scrn);

    src = (byte *)W_CacheLumpName(name, PU_CACHE);
    dest = scrn->data;

    for (y = 0; y < scrn->height; y++)
    {
        for (x = 0; x < scrn->width / 64; x++)
        {
            memcpy(dest, src + ((y & 63) << 6), 64);
            dest += 64;
        }
        if (scrn->width & 63)
        {
            memcpy(dest, src + ((y & 63) << 6), scrn->width & 63);
            dest += (scrn->width & 63);
        }
    }

    I_UnlockScreen(scrn);
}
