//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// Copyright 1993-1997 id Software
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// DESCRIPTION:
//	Handles WAD file header, directory, lump I/O.
//
//-----------------------------------------------------------------------------

#include <ctype.h>
#include <sys/types.h>
#include <string.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>
#if !defined(__WATCOMC__) && !defined(_MSC_VER) && !defined(__MINGW32__)
#include <alloca.h>
#define O_BINARY 0
#endif

#include "doomtype.h"
#include "m_swap.h"
#include "i_system.h"
#include "z_zone.h"
#include "w_wad.h"
#include "m_alloc.h"
#include "c_console.h"

#ifdef _MSC_VER
#include <io.h>
#else
#include <unistd.h>
#endif

//
// GLOBALS
//

// Location of each lump on disk.
lumpinfo_t *lumpinfo;
int numlumps;

void **lumpcache;

static void W_MergeBlocks(const char *start, const char *end, int namespace);

#if !defined(__WATCOMC__) && !defined(_MSC_VER) && !defined(__MINGW32__)

#define strcmpi strcasecmp

void strupr(char *s)
{
    while (*s)
    {
        *s = toupper(*s);
        s++;
    }
}

int filelength(int handle)
{
    struct stat fileinfo;

    if (fstat(handle, &fileinfo) == -1)
        I_Error("Error fstating");

    return fileinfo.st_size;
}

#endif

void ExtractFileBase(char *path,
                     char *dest)
{
    char *src;
    int length;

    src = path + strlen(path) - 1;

    // back up until a \ or the start
    while (src != path && *(src - 1) != '\\' && *(src - 1) != '/')
    {
        src--;
    }

    // copy up to eight characters
    memset(dest, 0, 8);
    length = 0;

    while (*src && *src != '.')
    {
        if (++length == 9)
            I_Error("Filename base of %s >8 chars", path);

        *dest++ = toupper((int)*src++);
    }
}

//
// LUMP BASED ROUTINES.
//

//
// W_AddFile
// All files are optional, but at least one file must be
//  found (PWAD, if all required lumps are present).
// Files with a .wad extension are wadlink files
//  with multiple lumps.
// Other files are single lumps with the base filename
//  for the lump name.
//

void W_AddFile(char *filename)
{
    wadinfo_t header;
    lumpinfo_t *lump_p;
    unsigned i;
    int handle;
    int length;
    int startlump;
    filelump_t *fileinfo;
    filelump_t singleinfo;

    // open the file and add to directory

    if ((handle = open(filename, O_RDONLY | O_BINARY)) == -1)
    {
        printf(" couldn't open %s\n", filename);
        return;
    }

    startlump = numlumps;

    if (strcmpi(filename + strlen(filename) - 3, "wad"))
    {
        // single lump file
        fileinfo = &singleinfo;
        singleinfo.filepos = 0;
        singleinfo.size = LONG(filelength(handle));
        ExtractFileBase(filename, singleinfo.name);
        numlumps++;
    }
    else
    {
        // WAD file
        read(handle, &header, sizeof(header));
        if (strncmp(header.identification, "IWAD", 4))
        {
            // Homebrew levels?
            if (strncmp(header.identification, "PWAD", 4))
            {
                I_Error("Wad file %s doesn't have IWAD "
                        "or PWAD id\n",
                        filename);
            }

            // ???modifiedgame = true;
        }
        header.numlumps = LONG(header.numlumps);
        header.infotableofs = LONG(header.infotableofs);
        length = header.numlumps * sizeof(filelump_t);
        fileinfo = (filelump_t *)alloca(length);
        lseek(handle, header.infotableofs, SEEK_SET);
        read(handle, fileinfo, length);
        numlumps += header.numlumps;
    }

    // Fill in lumpinfo
    if (!lumpinfo)
        lumpinfo = (lumpinfo_t *)M_Malloc(numlumps * sizeof(lumpinfo_t));
    else
        lumpinfo = (lumpinfo_t *)M_Realloc(lumpinfo, numlumps * sizeof(lumpinfo_t));

    if (!lumpinfo)
        I_Error("Couldn't realloc lumpinfo");

    lump_p = &lumpinfo[startlump];

    for (i = startlump; i < (unsigned)numlumps; i++, lump_p++, fileinfo++)
    {
        lump_p->handle = handle;
        lump_p->position = LONG(fileinfo->filepos);
        lump_p->size = LONG(fileinfo->size);
        lump_p->namespace = ns_global;
        strncpy(lump_p->name, fileinfo->name, 8);
    }
}

//
// W_InitMultipleFiles
// Pass a null terminated list of files to use.
// All files are optional, but at least one file
//  must be found.
// Files with a .wad extension are idlink files
//  with multiple lumps.
// Other files are single lumps with the base filename
//  for the lump name.
// Lump names can appear multiple times.
// The name searcher looks backwards, so a later file
//  does override all earlier ones.
//
void W_InitMultipleFiles (char **filenames, int numfiles)
{	
    int i;

    W_Clear();

    Printf("W_InitMultipleFiles: loading %d files\n", numfiles);
    
    // open all the files, load headers, and count lumps
    numlumps = 0;

    for (i = 0; i < numfiles; i++)
    {
        char *filename = filenames[i];

        printf(" adding %s\n", filename);
        W_AddFile (filename);
    }

    if (!numlumps)
        I_Error ("W_InitFiles: no files found");  
    
    W_MergeBlocks("S_START", "S_END", ns_sprites);
    W_MergeBlocks("F_START", "F_END", ns_flats);
    W_MergeBlocks("C_START", "C_END", ns_colormaps);
    
    W_InitCache();
}

//
// W_MergeBlocks
//
// Merges all given blocks into one superblock, allowing PWADs to properly
// define sprites and flats
// Adapted from Boom with very minor modifications
//
static int IsMarker(const char *marker, const char *name)
{
    return !strnicmp(name, marker, 8) || (*name == *marker && !strnicmp(name+1, marker, 7));
}

void W_MergeBlocks(const char *start, const char *end, int namespace)
{
    lumpinfo_t *marked;
    size_t i, nummarked = 0, numunmarked = 0;
    boolean ismarked = 0, markend = 0;
    lumpinfo_t *lump = lumpinfo;

    marked = (lumpinfo_t *)M_Malloc(sizeof(lumpinfo_t) * numlumps);

    for (i = numlumps; i--; lump++)
    {
        if (IsMarker(start, lump->name))
        {
            if (!nummarked)
            {
                strncpy(marked->name, start, 8);
                marked->size = 0;
                marked->namespace = ns_global;
                nummarked = 1;
            }
            ismarked = true;
        }
        else if (IsMarker(end, lump->name))
        {
            markend = true;
            ismarked = false;
        }
        else if (ismarked)
        {
            marked[nummarked] = *lump;
            marked[nummarked++].namespace = namespace;
        }
        else
            lumpinfo[numunmarked++] = *lump;
    }
    
    // Append marked list to end of unmarked list
    memcpy(lumpinfo + numunmarked, marked, nummarked * sizeof(*marked));

    M_Free(marked);

    numlumps = numunmarked + nummarked;

    if (markend)
    {
        lumpinfo[numlumps].size = 0;  // killough 3/20/98: force size to be 0
        lumpinfo[numlumps].namespace = ns_global;   // killough 4/17/98
        strncpy(lumpinfo[numlumps++].name, end, 8);
    }
}

void W_Clear()
{
    if (lumpinfo) M_Free(lumpinfo);
    if (lumpcache) M_Free(lumpcache);

    numlumps = 0;
    lumpinfo = NULL;
    lumpcache = NULL;

    Z_FreeTags(PU_CACHE, PU_CACHE);
}

void W_InitCache()
{
    int size;

    // set up caching
    size = numlumps * sizeof(*lumpcache);
    if (lumpcache)
        lumpcache = (void **)M_Realloc(lumpcache, size);
    else
        lumpcache = (void **)M_Malloc(size);

    if (!lumpcache)
        I_Error("Couldn't allocate lumpcache");

    memset(lumpcache, 0, size);
}

//
// W_NumLumps
//
int W_NumLumps(void)
{
    return numlumps;
}

//
// W_CheckNumForName
// Returns -1 if name not found.
//

int W_CheckNumForName(char *name)
{
    union {
        char s[9];
        int x[2];

    } name8;

    int v1;
    int v2;
    lumpinfo_t *lump_p;

    // make the name into two integers for easy compares
    strncpy(name8.s, name, 8);

    // in case the name was a fill 8 chars
    name8.s[8] = 0;

    // case insensitive
    strupr(name8.s);

    v1 = name8.x[0];
    v2 = name8.x[1];

    // scan backwards so patch lump files take precedence
    lump_p = lumpinfo + numlumps;

    while (lump_p-- != lumpinfo)
    {
        if (*(int *)lump_p->name == v1 && *(int *)&lump_p->name[4] == v2)
        {
            return lump_p - lumpinfo;
        }
    }

    // TFB. Not found.
    return -1;
}

//
// W_GetNumForName
// Calls W_CheckNumForName, but bombs out if not found.
//
int W_GetNumForName(char *name)
{
    int i;

    i = W_CheckNumForName(name);

    if (i == -1)
        I_Error("W_GetNumForName: %s not found!", name);

    return i;
}

//
// W_LumpLength
// Returns the buffer size needed to load the given lump.
//
int W_LumpLength(int lump)
{
    if (lump >= numlumps)
        I_Error("W_LumpLength: %i >= numlumps", lump);

    return lumpinfo[lump].size;
}

//
// W_ReadLump
// Loads the lump into the given buffer,
//  which must be >= W_LumpLength().
//
void W_ReadLump(int lump,
                void *dest)
{
    int c;
    lumpinfo_t *l;
    int handle;

    if (lump >= numlumps)
        I_Error("W_ReadLump: %i >= numlumps", lump);

    l = lumpinfo + lump;

    handle = l->handle;

    lseek(handle, l->position, SEEK_SET);
    c = read(handle, dest, l->size);

    if (c < l->size)
        I_Error("W_ReadLump: only read %i of %i on lump %i (err: %s)",
                c, l->size, lump, strerror(errno));

    if (l->handle == -1)
        close(handle);
}

unsigned W_ReadPartialLump(int lump, void *dest, unsigned offset, unsigned size)
{
    int c;
    lumpinfo_t *l;
    int handle;

    if (lump >= numlumps)
        I_Error("W_ReadLump: %i >= numlumps", lump);

    if (size == 0)
        return 0;

    l = lumpinfo + lump;

    handle = l->handle;

    if (size >= (unsigned)l->size)
        size = (unsigned)l->size;

    lseek(handle, l->position + offset, SEEK_SET);
    c = read(handle, dest, size);

    if (c == -1)
        I_Error("W_ReadLump: only read %i of %i on lump %i (err: %s)",
                c, size, lump, strerror(errno));

    if (l->handle == -1)
        close(handle);

    return (unsigned)c;
}

//
// W_CacheLumpNum
//
void *W_CacheLumpNum(int lump, int tag)
{
    byte *ptr;

    if (lump >= numlumps)
        I_Error("W_CacheLumpNum: %i >= numlumps", lump);

    if (!lumpcache[lump])
    {
        // read the lump in

        //printf ("cache miss on lump %i\n",lump);
        ptr = (byte *)Z_Malloc(W_LumpLength(lump), tag, &lumpcache[lump]);
        W_ReadLump(lump, lumpcache[lump]);
    }
    else
    {
        //printf ("cache hit on lump %i\n",lump);
        Z_ChangeTag(lumpcache[lump], tag);
    }

    return lumpcache[lump];
}

//
// W_CacheLumpName
//
void *
W_CacheLumpName(char *name,
                int tag)
{
    return W_CacheLumpNum(W_GetNumForName(name), tag);
}

//
// W_Profile
//
int info[2500][10];
int profilecount;

void W_Profile(void)
{
    int i;
    memblock_t *block;
    void *ptr;
    char ch;
    FILE *f;
    int j;
    char name[9];

    for (i = 0; i < numlumps; i++)
    {
        ptr = lumpcache[i];
        if (!ptr)
        {
            ch = ' ';
            continue;
        }
        else
        {
            block = (memblock_t *)((byte *)ptr - sizeof(memblock_t));
            if (block->tag < PU_PURGELEVEL)
                ch = 'S';
            else
                ch = 'P';
        }
        info[i][profilecount] = ch;
    }
    profilecount++;

    f = fopen("waddump.txt", "w");
    name[8] = 0;

    for (i = 0; i < numlumps; i++)
    {
        memcpy(name, lumpinfo[i].name, 8);

        for (j = 0; j < 8; j++)
            if (!name[j])
                break;

        for (; j < 8; j++)
            name[j] = ' ';

        fprintf(f, "%s ", name);

        for (j = 0; j < profilecount; j++)
            fprintf(f, "    %c", info[i][j]);

        fprintf(f, "\n");
    }
    fclose(f);
}
