//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// Copyright 1993-1997 id Software
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
//-----------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#define __BYTEBOOL__
#include <windows.h>
#include <d3d9.h>
#include <d3dx9math.h>
#include <time.h>
#include "i_win32.h"
#include "i_system.h"
#include "v_video.h"
#include "i_video.h"
#include "multires.h"
#include "doomtype.h"
#include "doomstat.h"
#include "w_wad.h"
#include "z_zone.h"
#include "version.h"
#include "c_console.h"
#include "m_argv.h"
#include "c_cvar.h"
#include "i_sound.h"

#define DOOMFVF (D3DFVF_XYZ | D3DFVF_TEX1)
#define GetScreenTexture(screen) ((LPDIRECT3DTEXTURE9)((screen)->platdata))

typedef struct 
{
    float X, Y, Z;
    float U, V;
} DoomVertex;

extern HWND win;

static LPDIRECT3D9 D3D;
static LPDIRECT3DDEVICE9 Device;
static LPDIRECT3DTEXTURE9 Texture; // Primary texture that gets rendered to the screen
static LPDIRECT3DTEXTURE9 PaletteTexture; // 256x1 RGBA texture containing the pallete
static LPDIRECT3DPIXELSHADER9 PaletteShader;
static LPDIRECT3DVERTEXBUFFER9 VertexBuffer;

static int fps = 0;
static int fpsFrames = 0;
static double lastFpsTime = 0.0;
static double DeltaTime = 0.0;
extern byte gammatable[256];

cvar_t *vid_showfps,
    *vid_vsync;

extern cvar_t *vid_fullscreen;

void I_InitD3D9()
{
    D3DPRESENT_PARAMETERS d3dpp;
    HRESULT hr;
    float sw;
    float sh;
    int i;
    UINT u;
    boolean supportsMode = 0;

    Printf("I_InitD3D9: Initialising Direct3D 9 backend.\n");

    sw = (float)SCREENWIDTH;
    sh = (float)SCREENHEIGHT;

    // This is required to avoid some weird alignment
    sw -= 0.025f;
    sh -= 0.025f;

    D3D = Direct3DCreate9(D3D_SDK_VERSION);
    if (D3D == NULL)
    {
        I_Error("Failed to create Direct3D 9.");
    }

    // Check whether fullscreen video mode is available
    if (vid_fullscreen->value)
    {
        D3DDISPLAYMODE mode;

        for (u = 0; u < IDirect3D9_GetAdapterModeCount(D3D, 0, D3DFMT_X8R8G8B8); u++)
        {
            IDirect3D9_EnumAdapterModes(D3D, 0, D3DFMT_X8R8G8B8, u, &mode);

            if (mode.Width == SCREENWIDTH && mode.Height == SCREENHEIGHT)
            {
                supportsMode = 1;
                break;
            }
        }

        if (!supportsMode)
        {
            I_Error("Display mode not supported.\n");
        }
    }

    ZeroMemory(&d3dpp, sizeof(D3DPRESENT_PARAMETERS));
    d3dpp.Windowed = !vid_fullscreen->value;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.hDeviceWindow = win;

    if (vid_vsync->value)
    {
        d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
        d3dpp.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
    }
    else
    {
        d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
        d3dpp.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
    }

    if (vid_fullscreen->value)
    {
        d3dpp.BackBufferCount = 1;
        d3dpp.BackBufferFormat = D3DFMT_X8R8G8B8;
        d3dpp.BackBufferWidth = SCREENWIDTH;
        d3dpp.BackBufferHeight = SCREENHEIGHT;
    }

    // Init the device
    hr = IDirect3D9_CreateDevice(D3D, D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, win, D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3dpp, &Device);
    if (hr != D3D_OK)
    {
        I_Error("Failed to create Direct3D 9 device.");
    }

    // Init the output texture
    hr = IDirect3DDevice9_CreateTexture(Device, SCREENWIDTH, SCREENHEIGHT, 1, D3DUSAGE_DYNAMIC, D3DFMT_L8, D3DPOOL_DEFAULT, &Texture, NULL);
    if (hr != D3D_OK)
    {
        I_Error("Failed to create Direct3D 9 texture.");
    }

    // Init the palette texture
    hr = IDirect3DDevice9_CreateTexture(Device, 256, 1, 1, D3DUSAGE_DYNAMIC, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &PaletteTexture, NULL);
    if (hr != D3D_OK)
    {
        I_Error("Failed to create Direct3D 9 palette texture.");
    }

    // Read in the shader bytecode and create the px shader
    {
        byte *code;
        int num, len;
        num = W_GetNumForName("ALDX9PS");
        len = W_LumpLength(num);
        code = (byte *)W_CacheLumpNum(num, PU_CACHE);

        hr = IDirect3DDevice9_CreatePixelShader(Device, (DWORD *)code, &PaletteShader);
        if (hr != D3D_OK)
        {
            I_Error("Failed to create Direct3D 9 pixel shader.");
        }
    }

    hr = IDirect3DDevice9_CreateVertexBuffer(Device, 6 * sizeof(DoomVertex), 0, DOOMFVF, D3DPOOL_MANAGED, &VertexBuffer, NULL);
    if (hr != D3D_OK)
    {
        I_Error("Failed to create Direct3D 9 vertex buffer.");
    }

    // Upload the vertices
    {
        DoomVertex verts[6] =
        {
            { 0.0f,		0.0f,		0.0f,		0.0f,		1.0f	},
            { 0.0f,		sh,			0.0f,		0.0f,		0.0f	},
            { sw,		sh,			0.0f,		1.0f,		0.0f	},
            { sw,		sh,			0.0f,		1.0f,		0.0f	},
            { sw,		0.0f,		0.0f,		1.0f,		1.0f	},
            { 0.0f,		0.0f,		0.0f,		0.0f,		1.0f	},
        };

        VOID *buf;
        hr = IDirect3DVertexBuffer9_Lock(VertexBuffer, 0, 0, (void **)&buf, 0);
        if (hr != D3D_OK)
        {
            I_Error("Failed to lock Direct3D 9 vertex buffer.");
        }

        memcpy(buf, verts, 6 * sizeof(DoomVertex));
        IDirect3DVertexBuffer9_Unlock(VertexBuffer);
    }

    // Set sampler state
    for (i = 0; i < 2; i++)
    {
        IDirect3DDevice9_SetSamplerState(Device, i, D3DSAMP_MINFILTER, D3DTEXF_POINT);
        IDirect3DDevice9_SetSamplerState(Device, i, D3DSAMP_MAGFILTER, D3DTEXF_POINT);
        IDirect3DDevice9_SetSamplerState(Device, i, D3DSAMP_MIPFILTER, D3DTEXF_NONE);
    }
}

void I_ShutdownD3D9()
{
    if (VertexBuffer)
        IDirect3DVertexBuffer9_Release(VertexBuffer);
    if (PaletteShader)
        IDirect3DPixelShader9_Release(PaletteShader);
    if (PaletteTexture)
        IDirect3DTexture9_Release(PaletteTexture);
    if (Texture)
        IDirect3DTexture9_Release(Texture);
    if (Device)
        IDirect3DDevice9_Release(Device);
    if (D3D)
        IDirect3D9_Release(D3D);
}

void I_AllocScreen(screen_t *screen, int width, int height, int depth)
{
    HRESULT hr;

    memset(screen, 0, sizeof(screen_t));

    screen->width = width;
    screen->height = height;
    screen->depth = depth;

    hr = IDirect3DDevice9_CreateTexture(Device, width, height, 1, D3DUSAGE_DYNAMIC, D3DFMT_L8, D3DPOOL_SYSTEMMEM, (LPDIRECT3DTEXTURE9 *)&screen->platdata, NULL);
    if (hr != D3D_OK)
    {
        I_Error("Failed to create Direct3D 9 texture.");
    }
}

void I_FreeScreen(screen_t *screen)
{
    LPDIRECT3DTEXTURE9 texture;

    while (screen->locked)
        I_UnlockScreen(screen);

    texture = GetScreenTexture(screen);

    IDirect3DTexture9_Release(texture);

    // zero it out
    memset(screen, 0, sizeof(screen_t));
}

boolean I_LockScreen(screen_t *screen)
{
    HRESULT hr;
    D3DLOCKED_RECT lockedrect;
    LPDIRECT3DTEXTURE9 texture;

    if (screen->locked)
    {
        screen->locked++;
        return 1;
    }

    if (screen == NULL || screen->platdata == NULL || screen->locked)
    {
        return 1;
    }

    texture = GetScreenTexture(screen);

    hr = IDirect3DTexture9_LockRect(texture, 0, &lockedrect, NULL, 0);
    if (hr != D3D_OK)
    {
        I_Error("Failed to lock Direct3D 9 palette texture.");
    }

    screen->data = (byte *)lockedrect.pBits;
    screen->pitch = lockedrect.Pitch;
    screen->locked++;

    return 1;
}

boolean I_UnlockScreen(screen_t *screen)
{
    LPDIRECT3DTEXTURE9 texture;

    if (!screen->locked)
    {
        return 1;
    }

    if (--screen->locked)
    {
        return 1;
    }

    texture = GetScreenTexture(screen);

    IDirect3DTexture9_UnlockRect(texture, 0);

    screen->data = NULL;
    screen->pitch = 0;
    screen->locked = 0;

    return 1;
}

void I_FinishUpdate(void)
{
    double start, end;
    
    I_DrawSoundOverlay();

    if (vid_showfps->value)
    {
        char str[128];
        int x, y;

        sprintf(str, "%d FPS (%f ms)", fps, DeltaTime);

        x = SCREENWIDTH - Con_StringWidth(str);
        y = 0;

        Con_DrawString(x, y, str);
    }

    start = I_GetMsTime();

    // Blit screens[0] to main texture
    IDirect3DDevice9_UpdateTexture(Device, (LPDIRECT3DBASETEXTURE9)GetScreenTexture(&screens[0]), (LPDIRECT3DBASETEXTURE9)Texture);

    // Set viewport
    {
        D3DVIEWPORT9 vp;
        vp.X = 0;
        vp.Y = 0;
        vp.Width = SCREENWIDTH;
        vp.Height = SCREENHEIGHT;
        vp.MinZ = 0.0f;
        vp.MaxZ = 1.0f;

        IDirect3DDevice9_SetViewport(Device, &vp);
    }

    // Is clearing even necessary?
    // We draw over the entire scene every frame and we have no other buffers to clear
    IDirect3DDevice9_Clear(Device, 0, NULL, D3DCLEAR_TARGET, 0xFF000000, 1.0f, 0);

    IDirect3DDevice9_BeginScene(Device);

    // Bind textures
    // Main screen is in sampler 0, palette in 1
    IDirect3DDevice9_SetTexture(Device, 0, (LPDIRECT3DBASETEXTURE9)Texture);
    IDirect3DDevice9_SetTexture(Device, 1, (LPDIRECT3DBASETEXTURE9)PaletteTexture);

    // Bind PS
    IDirect3DDevice9_SetPixelShader(Device, PaletteShader);

    // Set vertex buffer
    IDirect3DDevice9_SetVertexShader(Device, NULL);
    IDirect3DDevice9_SetStreamSource(Device, 0, VertexBuffer, 0, sizeof(DoomVertex));
    IDirect3DDevice9_SetFVF(Device, DOOMFVF);

    // Set projection matrix
    {
        D3DXMATRIX mat;
        D3DXMatrixOrthoOffCenterLH(&mat, 0.0f, (float)SCREENWIDTH, 0.0f, (float)SCREENHEIGHT, -1.0f, 1.0f);

        IDirect3DDevice9_SetTransform(Device, D3DTS_PROJECTION, &mat);
    }

    IDirect3DDevice9_DrawPrimitive(Device, D3DPT_TRIANGLELIST, 0, 2);

    IDirect3DDevice9_EndScene(Device);

    IDirect3DDevice9_Present(Device, NULL, NULL, NULL, NULL);

    end = I_GetMsTime();
    DeltaTime = end - start;
    fpsFrames++;
    if (end - lastFpsTime >= 1000.0)
    {
        fps = fpsFrames;
        fpsFrames = 0;
        lastFpsTime = end;
    }
}

void I_ReadScreen(byte *scr)
{
    I_LockScreen(&screens[0]);
    memcpy(scr, screens[0].data, screens[0].height * screens[0].pitch);
    I_UnlockScreen(&screens[0]);
}

void I_SetPalette(byte *pal)
{
    HRESULT hr;
    D3DLOCKED_RECT lockedrect;
    int i;
    byte *data;

    // don't set pallete if not initialised yet
    if (!PaletteTexture)
    {
        return;
    }

    hr = IDirect3DTexture9_LockRect(PaletteTexture, 0, &lockedrect, NULL, D3DLOCK_DISCARD);
    if (hr != D3D_OK)
    {
        I_Error("Failed to lock Direct3D 9 palette texture.");
    }

    data = (byte *)lockedrect.pBits;

    // Palette texture is in ARGB format
    for (i = 0; i < 256; i++)
    {
        byte *src = pal + (i * 3);
        byte *dst = data + (i * 4);

        dst[3] = 0xFF;
        dst[2] = gammatable[src[0]];
        dst[1] = gammatable[src[1]];
        dst[0] = gammatable[src[2]];
    }

    IDirect3DTexture9_UnlockRect(PaletteTexture, 0);
}

void I_Screenshot()
{
    RECT clientRect;
    int sw, sh, fw, fh, ch;
    LPDIRECT3DSURFACE9 surface;
    char filename[512];
    char timestamp[64];
    static char lastTimestamp[64];
    static unsigned counter = 0;

    counter++;

    // Generate filename
    {
        time_t rawtime;
        struct tm *info;
        time(&rawtime);
        info = localtime(&rawtime);

        strftime(timestamp, 64, "%Y-%m-%d_%H-%M-%S", info);

        if (strcmp(timestamp, lastTimestamp))
        {
            memcpy(lastTimestamp, timestamp, 64);
            counter = 0;
        }

        sprintf(filename, "%s/Screenshot_%s_%s_%u.png", I_GetConfigDir(), PACKAGE_ID, timestamp, counter);
    }

    // Calculate region to crop...
    GetWindowRect(win, &clientRect);
    sw = GetSystemMetrics(SM_CXSCREEN);
    sh = GetSystemMetrics(SM_CYSCREEN);
    fw = GetSystemMetrics(SM_CXFRAME);
    fh = GetSystemMetrics(SM_CYFRAME);
    ch = GetSystemMetrics(SM_CYCAPTION);
    
    clientRect.left += fw;
    clientRect.right -= fw;
    clientRect.top += fh + ch;
    clientRect.bottom -= fh;

    // Save the screenshot
    IDirect3DDevice9_CreateOffscreenPlainSurface(Device, sw, sh, D3DFMT_A8R8G8B8, D3DPOOL_SYSTEMMEM, &surface, NULL);
    IDirect3DDevice9_GetFrontBufferData(Device, 0, surface);
    D3DXSaveSurfaceToFileA(filename, D3DXIFF_PNG, surface, NULL, &clientRect);

    IDirect3DSurface9_Release(surface);

    Printf("Saved screenshot as %s\n", filename);
}
