//-----------------------------------------------------------------------------
//
// Copyright 2019-2020 Sean Baggaley
// Copyright 1993-1997 id Software
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its 
//    contributors may be used to endorse or promote products derived from this
//    software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
//-----------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shlobj.h>
#include <mmsystem.h>
#include <winsock.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <shlwapi.h>
#include "doomdef.h"
#include "m_misc.h"
#include "i_video.h"
#include "i_sound.h"
#include "d_net.h"
#include "g_game.h"
#include "i_system.h"
#include "m_alloc.h"
#include "version.h"

extern HWND win;

int mb_used = 6;

void I_Tactile(int on, int off, int total)
{
    // UNUSED.
    on = off = total = 0;
}

ticcmd_t emptycmd;
ticcmd_t *I_BaseTiccmd(void)
{
    return &emptycmd;
}

int I_GetHeapSize(void)
{
    return mb_used * 1024 * 1024;
}

byte *I_ZoneBase(int *size)
{
    void *block;

    *size = mb_used * 1024 * 1024;
    block = (byte *)M_Malloc(*size);
    memset(block, 0, *size);

    return (byte *)block;
}

static BOOL DirectoryExists(LPCTSTR szPath)
{
    DWORD dwAttrib = GetFileAttributes(szPath);

    return (dwAttrib != INVALID_FILE_ATTRIBUTES && (dwAttrib & FILE_ATTRIBUTE_DIRECTORY));
}

const char *I_GetConfigDir()
{
    static boolean inited = false;
    static char buffer[MAX_PATH];

    if (!inited)
    {
        memset(buffer, 0, sizeof(buffer));

        if (FAILED(SHGetFolderPath(NULL, CSIDL_APPDATA, NULL, 0, buffer)))
        {
            I_Error("Failed to find config directory\n");
        }

        buffer[strlen(buffer)] = '\\';
        strncat(buffer, PACKAGE_ID, sizeof(buffer));

        if (!DirectoryExists(buffer))
            if (!CreateDirectoryA(buffer, NULL))
                I_Error("Failed to create config directory %s\n", buffer);

        inited = true;
    }

    return buffer;
}

const char *I_GetExeDir()
{
    static boolean inited = false;
    static char buffer[MAX_PATH];

    if (!inited)
    {
        memset(buffer, 0, sizeof(buffer));

        GetModuleFileName(NULL, buffer, MAX_PATH);
        PathRemoveFileSpec(buffer);

        inited = true;
    }

    return buffer;
}

//
// I_GetTime
// returns time in 1/70th second tics
//
int I_GetTime(void)
{
    static DWORD startTime = 0;
    DWORD ms;
    int t;

    ms = timeGetTime();
    if (startTime == 0)
        startTime = ms;
    t = (int)(((ms - startTime) * TICRATE) / 1000);
    return t;
}

double I_GetMsTime()
{
    LARGE_INTEGER qpc, qpf;
    double v;

    QueryPerformanceCounter(&qpc);
    QueryPerformanceFrequency(&qpf);

    v = (((double)qpc.QuadPart) / ((double)qpf.QuadPart));
    v *= 1000.0;
    return v;
}

//
// I_Init
//
void I_Init(void)
{
    timeBeginPeriod(1);
    I_InitSound();
}

//
// I_Quit
//
void I_Quit(void)
{
    if (demorecording)
        G_CheckDemoStatus();

    D_QuitNetGame();
    I_ShutdownSound();
    M_SaveDefaults();
    I_ShutdownGraphics();

    WSACleanup();
    timeEndPeriod(1);

    exit(0);
}

void I_WaitVBL(int count)
{
    Sleep((1000 / 60) * count); /* [Petteri] Is this a good idea? */
}

byte *I_AllocLow(int length)
{
    byte *mem;

    mem = (byte *)M_Malloc(length);
    memset(mem, 0, length);
    return mem;
}

//
// I_Error
//
extern boolean demorecording;

void I_Error(char *error, ...)
{
    char buf[1024];
    va_list argptr;

    // Message first.
    va_start(argptr, error);
    vsnprintf(buf, sizeof(buf), error, argptr);
    va_end(argptr);

    fprintf(stderr, "Error: %s\n", buf);
    fflush(stderr);

    MessageBoxA(win, buf, "Fatal Error!", MB_OK|MB_ICONERROR);

    // Shutdown. Here might be other errors.
    if (demorecording)
        G_CheckDemoStatus();

    D_QuitNetGame();
    I_ShutdownGraphics();

    /* [Petteri] We'd better shut down the sound system as well: */
    I_ShutdownSound();

    WSACleanup();
    timeEndPeriod(1);

    exit(-1);
}

#define ALTAIR_KEYMAP(k, s) \
    case k: return s;

const char *I_MapKeyToName(event_t *event)
{
    static char ascii[2];
    BYTE kbd[256];
    int key, scancode;
    WORD c;

    switch (event->type)
    {
    case ev_keydown:
    case ev_keyup:
        {
            key = event->data2;
            scancode = event->data3;

            switch (key)
            {
            ALTAIR_KEYMAP(VK_BACK, "backspace");
            ALTAIR_KEYMAP(VK_TAB, "tab");
            ALTAIR_KEYMAP(VK_CLEAR, "clear");
            ALTAIR_KEYMAP(VK_RETURN, "enter");
            ALTAIR_KEYMAP(VK_SHIFT, "shift");
            ALTAIR_KEYMAP(VK_CONTROL, "ctrl");
            ALTAIR_KEYMAP(VK_MENU, "alt");
            ALTAIR_KEYMAP(VK_PAUSE, "pause");
            ALTAIR_KEYMAP(VK_ESCAPE, "escape");
            ALTAIR_KEYMAP(VK_SPACE, "space");
            ALTAIR_KEYMAP(VK_PRIOR, "pgup");
            ALTAIR_KEYMAP(VK_NEXT, "pgdn");
            ALTAIR_KEYMAP(VK_END, "end");
            ALTAIR_KEYMAP(VK_HOME, "home");
            ALTAIR_KEYMAP(VK_LEFT, "left");
            ALTAIR_KEYMAP(VK_UP, "up");
            ALTAIR_KEYMAP(VK_RIGHT, "right");
            ALTAIR_KEYMAP(VK_DOWN, "down");
            ALTAIR_KEYMAP(VK_SELECT, "select");
            ALTAIR_KEYMAP(VK_SNAPSHOT, "printscr");
            ALTAIR_KEYMAP(VK_INSERT, "insert");
            ALTAIR_KEYMAP(VK_DELETE, "delete");
            ALTAIR_KEYMAP(VK_F1, "f1");
            ALTAIR_KEYMAP(VK_F2, "f2");
            ALTAIR_KEYMAP(VK_F3, "f3");
            ALTAIR_KEYMAP(VK_F4, "f4");
            ALTAIR_KEYMAP(VK_F5, "f5");
            ALTAIR_KEYMAP(VK_F6, "f6");
            ALTAIR_KEYMAP(VK_F7, "f7");
            ALTAIR_KEYMAP(VK_F8, "f8");
            ALTAIR_KEYMAP(VK_F9, "f9");
            ALTAIR_KEYMAP(VK_F10, "f10");
            ALTAIR_KEYMAP(VK_F11, "f11");
            ALTAIR_KEYMAP(VK_F12, "f12");
            ALTAIR_KEYMAP(VK_F13, "f13");
            ALTAIR_KEYMAP(VK_F14, "f14");
            ALTAIR_KEYMAP(VK_F15, "f15");
            ALTAIR_KEYMAP(VK_F16, "f16");
            ALTAIR_KEYMAP(VK_F17, "f17");
            ALTAIR_KEYMAP(VK_F18, "f18");
            ALTAIR_KEYMAP(VK_F19, "f19");
            ALTAIR_KEYMAP(VK_F20, "f20");
            ALTAIR_KEYMAP(VK_F21, "f21");
            ALTAIR_KEYMAP(VK_F22, "f22");
            ALTAIR_KEYMAP(VK_F23, "f23");
            ALTAIR_KEYMAP(VK_F24, "f24");
            ALTAIR_KEYMAP(VK_OEM_8, "`");
            }

            // return the ASCII name if possible
            GetKeyboardState(kbd);

            if (ToAscii(key, scancode, kbd, &c, 0) == 1)
            {
                ascii[0] = (char)c;
                ascii[1] = 0;

                return ascii;
            }

            return NULL;
        }
        break;
    }
    
    return NULL;
}
